Package: spiky
Type: Package
Title: Spike-in calibration for cell-free MeDIP
Description: spiky implements methods and model generation for cfMeDIP (cell-free methylated DNA immunoprecipitation) with spike-in controls. CfMeDIP is an enrichment protocol which avoids destructive conversion of scarce template, making it ideal as a "liquid biopsy," but creating certain challenges in comparing results across specimens, subjects, and experiments. The use of synthetic spike-in standard oligos allows diagnostics performed with cfMeDIP to quantitatively compare samples across subjects, experiments, and time points in both relative and absolute terms.
Version: 1.19.0
Date: 2023-04-19
Authors@R: c(person("Samantha", "Wilson", role=c("aut")), 
             person("Lauren", "Harmon", role=c("aut")),
             person("Tim", "Triche", role=c("aut","cre"),
                    email="trichelab@gmail.com"))
biocViews: DifferentialMethylation, DNAMethylation, Normalization,
        Preprocessing, QualityControl, Sequencing
URL: https://github.com/trichelab/spiky
BugReports: https://github.com/trichelab/spiky/issues
License: GPL-2
Depends: Rsamtools, GenomicRanges, R (>= 3.6.0)
Imports: stats, scales, bamlss, methods, tools, IRanges, Biostrings,
        GenomicAlignments, BlandAltmanLeh, GenomeInfoDb, BSgenome,
        S4Vectors, graphics, ggplot2, utils
Suggests: covr, testthat, rmarkdown, markdown, knitr, devtools,
        BSgenome.Mmusculus.UCSC.mm10.masked,
        BSgenome.Hsapiens.UCSC.hg38.masked, BiocManager
RoxygenNote: 7.2.1
Roxygen: list(markdown = TRUE)
VignetteBuilder: knitr
Encoding: UTF-8
LazyData: true
git_url: https://git.bioconductor.org/packages/spiky
git_branch: devel
git_last_commit: 42d312e
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-05-18
NeedsCompilation: no
Packaged: 2026-05-18 23:33:27 UTC; biocbuild
Author: Samantha Wilson [aut],
  Lauren Harmon [aut],
  Tim Triche [aut, cre]
Maintainer: Tim Triche <trichelab@gmail.com>
Built: R 4.6.0; ; 2026-05-19 05:10:51 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_frag_info)
export(bam_to_bins)
export(bin_pmol)
export(convertPairedGRtoGR)
export(covg_to_df)
export(find_spike_contigs)
export(generate_spike_fasta)
export(get_base_name)
export(get_binned_coverage)
export(get_merged_gr)
export(get_spike_depth)
export(get_spiked_coverage)
export(kmax)
export(kmers)
export(methylation_specificity)
export(model_bam_standards)
export(model_glm_pmol)
export(parse_spike_UMI)
export(predict_pmol)
export(process_spikes)
export(read_bedpe)
export(rename_spike_seqlevels)
export(rename_spikes)
export(scan_genomic_bedpe)
export(scan_genomic_contigs)
export(scan_methylation_specificity)
export(scan_spike_bedpe)
export(scan_spike_contigs)
export(scan_spike_counts)
export(scan_spiked_bam)
export(seqinfo_from_header)
export(spike_bland_altman_plot)
export(spike_counts)
export(tile_bins)
exportMethods(plot)
import(BSgenome)
import(Biostrings)
import(BlandAltmanLeh)
import(GenomeInfoDb)
import(GenomicAlignments)
import(GenomicRanges)
import(IRanges)
import(Rsamtools)
import(S4Vectors)
import(bamlss)
import(ggplot2)
import(methods)
import(scales)
import(tools)
importFrom(graphics,barplot)
importFrom(graphics,par)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(stats,glm)
importFrom(stats,predict)
importFrom(utils,data)

[![Spiky](vignettes/spiky-small.png)](https://github.com/trichelab/spiky)

# spiky

[![Build Status](https://travis-ci.org/trichelab/spiky.png?branch=master)](https://travis-ci.org/trichelab/spiky)  [![codecov](https://codecov.io/gh/trichelab/spiky/branch/master/graph/badge.svg)](https://codecov.io/gh/trichelab/spiky)

# Spike-ins for everyone 
Analysis of cfMeDIP-seq data with spike-in control sequences. For more information, see Wilson, et. al, Sensitive and reproducible cell-free methylome quantification with synthetic spike-in controls, Cell Reports Methods, 2022. (https://www.sciencedirect.com/science/article/pii/S266723752200176X?via%3Dihub)

Please note that `spiky` strongly suggests a recent version of R (4.0 or newer; 3.6 or newer is required).

This restriction may be removed if our dependencies can tolerate it.

## Installing

To install this package, start R (version "3.6" or later) and enter:

```
if (!requireNamespace("BiocManager", quietly = TRUE)) { 
  install.packages("BiocManager")
}
BiocManager::install("trichelab/spiky")
```

## For developers

The repository includes a Makefile to facilitate some common tasks.

### Running tests

`$ make test`. Requires the [testthat](https://github.com/hadley/testthat) package. You can also specify a specific test file or files to run by adding a "file=" argument, like `$ make test file=logging`. `test_package` will do a regular-expression pattern match within the file names. See its documentation in the `testthat` package.

### Updating documentation

`$ make doc`. Requires the [roxygen2](https://github.com/klutometis/roxygen) package.

# This script generates the object data/testGR.rda
#
# column 1 (UMI) is contigUMI5_UMI5 where each UMI5 is from one read in the pair
# seq is fragment sequence after adapter trimming and UMI decoding
# rname is the contig fragment of the UMI variable
# pos is position along the unplaced/MT contig in hg38
# 6547 is one sample
# 6548 is another
#
testfile <- system.file("extdata","2020_Human0.01_test.csv.gz",package="spiky")
test <- read.table(testfile, sep=" ", head=TRUE) # bit mysterious, see above
# really only need UMI, pos, seq, read_count for each subject/UMI combo
testGR <- parse_spike_UMI(test$UMI, pos=test$pos, seqs=test$seq)
# how can we make it so that spiky:::.addFragInfo() works on this?
# at present, .addFragInfo needs a `frag_grp` column, and test.csv lacks one
message("Ideally, we would like to have frag_grp information in testGR")
save(testGR, file="../../data/testGR.rda")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(devtools)
load_all("./")

## ----eval = FALSE,message=FALSE-----------------------------------------------
# #To install this package, start R (version "3.6" or later) and enter:
#   #if (!requireNamespace("BiocManager", quietly = TRUE))
#   #  install.packages("BiocManager")
#   #
#   #BiocManager::install("spiky")
# 
# library(spiky)

## -----------------------------------------------------------------------------
data(spike)

## -----------------------------------------------------------------------------
sb <- system.file("extdata", "example.spike.bam", package="spiky",              mustWork=TRUE)
outFasta <- paste(system.file("extdata", package="spiky", mustWork=TRUE),"/spike_contigs.fa",sep="")
show(generate_spike_fasta(sb, spike=spike,fa=outFasta))

## -----------------------------------------------------------------------------
spikes <- system.file("extdata", "spikes.fa", package="spiky", mustWork=TRUE)
spikemeth <- spike$methylated
process_spikes(spikes, spikemeth)

## ----eval=TRUE----------------------------------------------------------------
genomic_bam_path <- system.file("extdata", "example_chr21.bam", package="spiky", mustWork=TRUE)
genomic_coverage <- scan_genomic_contigs(genomic_bam_path,spike=spike)
spike_bam_path <- system.file("extdata", "example.spike.bam", package="spiky", mustWork=TRUE)
spikes_coverage <- scan_spike_contigs(spike_bam_path,spike=spike)


## -----------------------------------------------------------------------------
genomic_bedpe_path <- system.file("extdata", "example_chr21_bedpe.bed.gz", package="spiky", mustWork=TRUE)
genomic_coverage <- scan_genomic_bedpe(genomic_bedpe_path,genome="hg38")
spike_bedpe_path <- system.file("extdata", "example_spike_bedpe.bed.gz", package="spiky", mustWork=TRUE)
spikes_coverage <- scan_spike_bedpe(spike_bedpe_path,spike=spike)

## ----eval=TRUE----------------------------------------------------------------
##Calculate methylation specificity
methyl_spec <- methylation_specificity(spikes_coverage,spike=spike)
print(methyl_spec)

## ----eval=TRUE----------------------------------------------------------------
## Build the Gaussian generalized linear model on the spike-in control data
gaussian_glm <- model_glm_pmol(spikes_coverage,spike=spike)
summary(gaussian_glm)

## ----eval=TRUE----------------------------------------------------------------
# Predict pmol concentration
# To select a genome other than hg38, use BSgenome::available.packages() to find valid BSgenome name
#library("BSgenome.Hsapiens.UCSC.hg38")
sample_data_pmol <- predict_pmol(gaussian_glm, genomic_coverage,bsgenome="BSgenome.Hsapiens.UCSC.hg38",ret="df")
head(sample_data_pmol,n=1)


## ----eval=TRUE----------------------------------------------------------------
sample_binned_data <- bin_pmol(sample_data_pmol)
head(sample_binned_data,n=1)

## -----------------------------------------------------------------------------
sessionInfo()

# This script turns a bunch of BED files into a GRanges with a matrix hanging off of it.

library(rtracklayer)

# in conjunction with tabixToGRanges, this can reduce arbitrarily many BEDs
mergeMcols <- function(x, y) {
  stopifnot (identical(granges(x), granges(y)))
  mcols(x)[, names(mcols(y))] <- mcols(y)
  return(x)
}

BEDs <- list.files(patt="bed.gz$")
names(BEDs) <- vapply(strsplit(BEDs, "\\."), `[`, 3, FUN.VALUE=character(1))

GRs <- lapply(BEDs, import)
for (samp in names(GRs)) names(mcols(GRs[[samp]]))[1] <- samp

merged <- Reduce(mergeMcols, GRs)
save(merged, file="../../Human01_pmol_test.rda")

# This script is used to retrieve mitochondrial gene sequences from GenBank.

data("genbank_mito", package="spiky")
show(genbank_mito)

library(genbankr)
get_gbas <- function(y) lapply(lapply(y, GBAccession), readGenBank)
GBAs <- get_gbas(genbank_mito$accession)
names(GBAs) <- vapply(GBAs, function(gba) seqnames(seqinfo(gba)), character(1))

get_seqs <- function(x) DNAStringSet(lapply(lapply(x, getSeq), `[[`, 1))
genbank_mito$sequence <- get_seqs(GBAs)
mtSpikes <- process_spikes(genbank_mito)

mtTxDbs <- lapply(GBAs, makeTxDbFromGenBank)
mtGenes <- lapply(mtTxDbs, genes)

get_mt_tx_seqs <- function(gba) {
  gxs <- genes(gba)
  names(gxs) <- gxs$gene
  mtseq <- getSeq(gba)
  mtgxs <- granges(gxs)
  mtgxs$gene <- gxs$gene
  mtgxs$sequence <- getSeq(mtseq, gxs)
  mtgxs
}

mtGeneSeqs <- lapply(GBAs, get_mt_tx_seqs)
mtTxSpikes <- lapply(mtGeneSeqs, process_spikes)

# This script is used to generate "data/dedup.rda" and "data/spike_read_counts.rda"
# load
dedup <- read.table("2020_dedup_spikein_data.csv.gz", sep=" ")[, c(1, 5, 6)]
# frag_grp has all information necessary to reconstruct the rest, so
save(dedup, file="../../data/dedup.rda")

# melt
library(reshape2)
spike_read_counts <- melt(dedup, id.vars="frag_grp")
colnames(spike_read_counts) <- sub("^variable$", "id",
                                   colnames(spike_read_counts))
colnames(spike_read_counts) <- sub("^value$", "read_count",
                                   colnames(spike_read_counts))
spike_read_counts$id <- as.factor(sub("read_count_", "run",
                                      spike_read_counts$id))
save(spike_read_counts, file="../../data/spike_read_counts.rda")

#' pretty much what it says: scan standard chroms + spike contigs from a BAM
#'
#' Note: behind the scenes, this is being refactored into scan_spike_contigs 
#' and scan_genomic_contigs. Once that is done, perhaps before release, the 
#' default workflow will switch to 
#' 
#' 1. scan spike contigs and count fragments per contig or per bin.
#' 2. fit the appropriate model for adjusting genomic contigs based on spikes.
#' 3. scan and adjust binned fragment tallies along genomic contigs per above.
#' 
#' This approach decouples binning schemes from model generation (using spikes) 
#' and model-based adjustment (using genomic fragment counts), decreasing code
#' complexity while increasing the opportunities for caching & parallelization.
#' 
#' @param bam       the BAM file
#' @param spike     the spike-in reference database (e.g. data(spike))
#' @param mapq      minimum mapq value to count a pair (20)
#' @param binwidth  width of the bins for chromosomal tiling (300)
#' @param bins      a pre-tiled GRanges for binning coverage (NULL)
#' @param how       how to record spike read coverage (max or mean)? (max)
#' @param dupe      unique (FALSE), duplicte (TRUE), or all (NA) reads? (FALSE)
#' @param paired    restrict coverage to that from properly paired reads? (TRUE)
#' @param standard  restrict non-spike contigs to "standard" chromosomes? (TRUE)
#' @param ...       additional arguments to pass to scanBamFlag()
#'
#' @return          a CompressedGRangesList with bin- and spike-level coverage
#'
#' @examples
#' library(GenomicRanges)
#' data(spike, package="spiky")
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' res <- scan_spiked_bam(sb, spike=spike, bins=GRanges())
#' summary(res$spikes$coverage)
#'
#' @details
#'   For a more realistic example (not run), one might do something like:
#'
#'   data(spike, package="spiky");
#'   bam <- "2021_ctl.hg38_withSpikes.bam";
#'   ssb_res <- scan_spiked_bam(bam, mapq=20, spike=spike);
#'
#'   An extract from the resulting `ssb_res` object is available via
#'
#'   data(ssb_res, package="spiky");
#'
#'   The full ssb_res is a GRangesList object with 300bp-binned coverage on the
#'   standard (chr1-22, chrX, chrY, chrM) chromosomes (as determined by the
#'   GenomeInfoDb::standardChromosomes() function against the assembly defined
#'   in the BAM or CRAM file, by default; if desired, a user can scan all
#'   genomic contigs by setting standard=FALSE when calling the function).
#'   By default, the mean base-level coverage of genomic bins is reported,
#'   and the maximum spike-level coverage is reported, though this can also
#'   be adjusted as needed. The results then inform the reliability of
#'   measurements from replicate samples in multiple labs, as well as the
#'   adjusted quantitative coverage in each bin once the absolute quantity
#'   of captured cell-free methylated DNA has been fit by model_glm_pmol and
#'   predict_pmol. In some sense, this function converts BAMs/CRAMs into usable
#'   data structures for high-throughput standardized cfMeDIP experiments.
#'
#'   The data extract used in other examples is the same as the full version,
#'   with the sole difference being that genomic bins are limited to chr22.
#'
#' @seealso         GenomeInfoDb::keepStandardChromosomes
#' @seealso         Rsamtools::ScanBamParam
#'
#' @import          GenomicAlignments
#' @import          GenomeInfoDb
#' @import          Rsamtools
#'
#' @export
scan_spiked_bam <- function(bam, spike, mapq=20, binwidth=300L, bins=NULL, how=c("max", "mean"), dupe=FALSE, paired=TRUE, standard=TRUE, ...) {

  # scan the BAM (or CRAM if supported) to determine which reads to import
  si <- seqinfo_from_header(bam)
  spikes <- find_spike_contigs(si, spike=spike)
  if (length(spikes) > 0) {
    mappings <- attr(spikes, "mappings")
    attr(spikes, "mappings") <- NULL
    genome(si)[spikes] <- "spike"
  } else {
    stop(bam, " doesn't appear to have any spike-ins among its contigs.")
  }

  # properly indexed
  bf <- BamFile(bam)

  # get a GRanges corresponding to our contig selections for the BAM/CRAM file
  gr <- get_merged_gr(si=si, spike=spike, standard=standard) # scanBam positive selections

  # tabulate spike coverage onto this subset of the filter GRanges from above
  spike_gr <- subset(gr, genome(gr)[as(seqnames(gr), "character")] == "spike")

  # create appropriate filters for coverage tabulation
  fl <- scanBamFlag(isDuplicate=dupe, isPaired=paired, isProperPair=paired, ...)
  bp <- ScanBamParam(flag=fl, which=gr)
  bamMapqFilter(bp) <- mapq

  # assess coverage on contigs we care about (bin the Rles later)
  # FIXME: this is the bottleneck in the code 
  covg <- get_spiked_coverage(bf=bf, bp=bp, gr=gr)

  # bins for coverage
  if (is.null(bins)) bins <- tile_bins(gr=gr, binwidth=binwidth)

  # genomic coverage is averaged across each bin
  if (length(bins) > 0) {
    message("Binning genomic coverage...")
    binned <- get_binned_coverage(bins, covg)
    message("Done.")
  } else {
    message("Empty bins provided, skipping genomic binning.")
    binned <- bins
  }

  # spikes get summarized as max or mean across each
  spiked <- get_spike_depth(covg, spike_gr, how=how, spike=spike)

  # toss them into a list of summary results
  res <- GRangesList(genomic=binned, spikes=spiked) # redundant mcols :-(
  return(res)

}

#' Build a generalized linear model from spike-ins to correct bias in cfMeDIP
#'
#' formerly '2020_model_glm_fmol'.  Note that everything in x can be had from
#' a BAM/CRAM with spike contigs named as frag_grp (len_CpGs_GC) in the index
#' and in fact that is what scan_spiked_bam now does.
#'
#' @param   x     data w/frag_grp, id, and read_count; or scan_spiked_bam result
#' @param   conc  concentration for each spike (will be referenced if NULL)
#' @param   spike spike database, e.g. `data(spike, package='spiky')`
#' @param   ...   other arguments to pass to `glm` (e.g. `family`)
#'
#' @return        the model fit for the data
#'
#' @examples
#'
#' data(spike, package="spiky")
#'
#' data(spike_read_counts, package="spiky")
#' fit1 <- model_glm_pmol(spike_read_counts, spike=spike)
#'
#' data(spike_res) # scan_spiked_bam result
#' fit2 <- model_glm_pmol(spike_res, spike=spike)
#'
#' @importFrom stats glm
#'
#' @export
model_glm_pmol <- function(x, spike, conc=NULL, ...) {

  if (!is(x, "data.frame")) {
    message("Converting x (a coverage result?) to a data.frame...")
    x <- covg_to_df(x, spike=spike)
  }

  cols <- c("frag_grp", "id", "read_count")
  if (!all(cols %in% names(x))) {
    stop("Input data must have columns frag_grp, id, and read_count")
  }

  x <- add_frag_info(x, "frag_grp") # now exported
  x$conc <- .getConcFromFraglen(x$fraglen,conc=conc, spike=spike) # below

  # Adjust for the 0.01ng dilution
  x$conc <- x$conc * 0.9

  # Cube root CpG number to normalize
  x$CpG <- as.integer(x$CpG) # this too would be helpful to standardize
  x$CpG_3 <- x$CpG ^ (1/3)

  # Gaussian model by default -- can alter with ... params
  fit <- glm(formula = conc ~ read_count + fraglen + GC + CpG_3, data = x, ...)
  r2_gaussian= 1 - (fit$deviance / fit$null.deviance)
  attr(fit, "r2_gaussian") <- r2_gaussian #  report in summary
  attr(fit, "data") <- x
  return(fit)

}


# helper fn
.getConcFromFraglen <- function(fraglen, spike, conc, concs=NULL) {

  # pull these from spike
  if (is.null(concs)) {
    tbl <- with(spike, table(width(sequence), conc))
    concs <- as.numeric(colnames(tbl)[max.col(tbl)])
    names(concs) <- rownames(tbl)
  }

  # show(concs)
  #    80   160   320
  # 0.004 0.002 0.001

  res <- concs[as.character(fraglen)]

  # essentially the fallthrough from ifelse
  res[is.na(res)] <- concs[length(concs)]

  return(res)

}

#' predict picomoles of DNA from a fit and read counts (coverage)
#'
#' FIXME: this could be made MUCH faster by precomputing CpG/GC stats per bin
#'
#' @param   fit       result of model_glm_pmol
#' @param   genomic_gr  the genomic data / new data
#' @param   bsgenome  BSgenome name (if null, will guess from genomic_gr)
#' @param   ret       return a data.frame ("df") or GRanges ("gr")?  ("gr")
#' @param   slide     compute a sliding window estimate for GCfrac (1/3 width)?
#'
#' @details
#' Using GRanges as the return value is (perhaps counterintuitively) *much*
#' faster than the data.frame, since the sequence of the bins gets converted
#' from a BSgenome representation to characters in the latter (it is implied
#' by the bin start, stop, and genome when left as a GRanges).
#'
#' @return  object with read count, fraglen, GC%, CpG**(1/3), and concentration
#'
#' @import tools
#' @import BSgenome
#' @import S4Vectors
#' @import Biostrings
#' @importFrom stats predict
#'
#' @examples
#'
#' data(spike_res)
#' data(genomic_res)
#' data(spike, package="spiky")
#' fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
#' preddf <- predict_pmol(fit, genomic_res, ret="df")
#' pred <- predict_pmol(fit, genomic_res, ret="gr")
#' bin_pmol(pred)
#'
#' @export
predict_pmol <- function(fit, genomic_gr, bsgenome=NULL, ret=c("gr", "df"), slide=FALSE) {

  # what shall be returned?  (eventually it would be nice to pass GRs around)
  ret <- match.arg(ret)

  # guess the genome if not provided (it won't be)
  assembly=bsgenome
  if (is.null(bsgenome)) {
    assembly <- grep("(hg|GRCh)", unique(genome(genomic_gr)), value=TRUE)
    genomepattern <- paste0(assembly, "$")
    bsgenome <- grep(genomepattern, available.genomes(), value=TRUE)
  }
  if (!bsgenome %in% installed.genomes()) {
    message("Your reads appear to be aligned against ", assembly)
    message("In order to correct for GC% and CpG density, you need a genome.")
    message("Our best guess for that genome is ", bsgenome)
    message("It doesn't look like you have installed it yet.")
    message("Try `BiocManager::install(\"", bsgenome, "\").)")
    stop("Need ", bsgenome, " to predict bin-level concentrations. Exiting.")
  }

  # Get read_count, fraglen, GC, and CpG_3 from GRanges object
  if (is.null(genomic_gr)) genomic_gr <- attr(fit, "data") # if feasible... ?
  reads <- subset(as(genomic_gr,"GRanges"), coverage != 0) # is this sensible (bias)?
  names(mcols(reads)) <- sub("coverage", "read_count", names(mcols(reads)))
  reads$fraglen <- width(reads)

  # can this be made superfluous?
  message("Adjusting for bin-level biases...")
  myGenome <- .load_genome(bsgenome)
  seqlengths(reads) <- seqlengths(myGenome)[names(genome(reads))]
  myBinSeqs <- Views(myGenome, reads)
  myBinGCs <- alphabetFrequency(myBinSeqs)[, c("G","C")]
  reads$GC <- rowSums(myBinGCs) / width(reads)
  myBinCpGs <- dinucleotideFrequency(myBinSeqs)[, "CG"]
  reads$CpG_3 <- myBinCpGs ** (1/3)
  names(reads) <- as.character(reads) # coords
  message("Done.")

  message("Starting pmol prediction...")
  reads$pred_conc <- predict(fit, as.data.frame(mcols(reads)))
  message("Done.")

  if (ret == "gr") return(reads)
  else return(.gr_prediction_to_df(reads, myBinSeqs))

}


# helper fn
.load_genome <- function(bsgenome) {
  if (!any(grepl(bsgenome, loadedNamespaces()))) {
    message("Attempting to load ", bsgenome, "...")
    x <- try(suppressMessages(attachNamespace(bsgenome)), silent=TRUE)
    if (inherits(x, "try-error")) stop("Unable to load library ", bsgenome, "!")
    message("OK.")
  }
  g <- try(getBSgenome(bsgenome, masked=FALSE, load.only=FALSE))
  if (inherits(g, "try-error")) stop("Unable to load ", bsgenome, ", exiting")
  return(g)
}


# helper fn
.gr_prediction_to_df <- function(gr, seqs) {

  res <- data.frame(chrom=decode(seqnames(gr)),
                    range.start=start(gr),
                    range.end=end(gr),
                    sequence=DNAStringSet(seqs),
                    read_count=gr$read_count,
                    fraglen=gr$fraglen,
                    GC=gr$GC,
                    CpG_3=gr$CpG_3,
                    pred_conc=gr$pred_conc)
  return(res)

}

#' create seqinfo (and thus a standard chromosome filter) from a BAM header
#'
#' @param   x     the BAM file or its header
#' @param   gen   genome of the BAM file, if known (NULL; autodetect)
#' @param   std   standard chromosomes only? (FALSE; will be empty if spikes)
#' @param   ret   return Seqinfo ("si", the default) or GRanges ("gr")? ("si")
#'
#' @return        Seqinfo object or GRanges (or `as(seqinfo, "GRanges")`)
#'
#' @details
#'
#'   Setting std=TRUE on a spike-in BAM will produce an empty result.
#'
#' @examples
#'
#'   library(Rsamtools)
#'   fl <- system.file("extdata", "ex1.bam", package="Rsamtools", mustWork=TRUE)
#'
#'   hdr <- scanBamHeader(BamFile(fl))
#'   si <- seqinfo_from_header(hdr)
#'   gr <- seqinfo_from_header(fl, ret="gr")
#'   stopifnot(identical(gr, as(si, "GRanges")))
#'
#'   std_si <- seqinfo_from_header(fl, std=TRUE)
#'   seqlevels(std_si)
#'
#'   # for comparison with below
#'   data(spike, package="spiky")
#'   spike
#'
#'   sp <- system.file("extdata", "example.spike.bam", package="spiky")
#'   sp_gr <- seqinfo_from_header(sp, ret="gr")
#'   sp_gr
#'
#' @import methods
#' @import Rsamtools
#' @import GenomeInfoDb
#'
#' @export
seqinfo_from_header <- function(x, gen=NA, std=FALSE, ret=c("si", "gr")) {

  ret <- match.arg(ret)

  if (is.list(x) & "targets" %in% names(x)) { # it's a header, just convert
    res <- .seqinfo_from_targets(x$targets, genome=gen)
  } else {
    if (!is(x, "BamFile")) x <- BamFile(x)
    res <- .seqinfo_from_targets(scanBamHeader(x)$targets, genome=gen)
  }

  if (std) res <- keepStandardChromosomes(res)
  if (ret == "gr") res <- as(res, "GRanges")
  return(res)

}


# helper fn -- may break this out eventually
.seqinfo_from_targets <- function(chrs, genome=NA) {

  if (is.null(names(chrs))) stop("Seqlengths have no contig names. Exiting.")
  if (is.list(chrs) & "targets" %in% names(chrs)) chrs <- chrs$targets
  res <- as(data.frame(chrom=names(chrs), seqlengths=chrs), "Seqinfo")

  if (!is.na(genome) & !is.null(genome)) {

    genome(res) <- genome

  } else {

    # we can deduce a bit here, at least for mouse and human:
    if ("chrM" %in% seqlevels(res) | "MT" %in% seqlevels(res)) {
      orig <- seqlevelsStyle(res)
      seqlevelsStyle(res) <- "UCSC"
      if ("chrM" %in% seqlevels(res)) {
        if (seqlengths(res)["chrM"] == 16299) genome(res) <- "mm10"
        if (seqlengths(res)["chrM"] == 16571) genome(res) <- "hg19" # boooo
        if (seqlengths(res)["chrM"] == 16569) {
          if ("chr1" %in% seqlevels(res) &
              seqlengths(res)["chr1"] == 248956422) {
            genome(res) <- "hg38"
          } else {
            genome(res) <- "GRCh37"
          }
        }
      }
      seqlevelsStyle(res) <- orig
    }

  }

  return(res)

}

#' compute methylation specificity for spike-in standards
#'
#' In a cfMeDIP experiment, the yield of methylated fragments should be >95%
#' (ideally 98-99%) due to the nature of the assay.
#'
#' @param   spike_gr    GRanges of spike contigs (e.g. output object from scan_spiked_bam, scan_spike_contigs, or scan_spike_bedpe)
#' @param   spike       spike contig database, if needed (e.g. data(spike))
#'
#' @return              list with median and mean coverage across spike contigs
#'
#' @examples
#' data(genomic_res)
#' data(spike_res)
#' data(spike, package="spiky")
#' methylation_specificity(spike_res, spike=spike)
#'
#' @export
methylation_specificity <- function(spike_gr, spike) {
  methreads <- covg_to_df(spike_gr, meth=TRUE, spike=spike)$read_count
  totalreads <- covg_to_df(spike_gr, meth=FALSE, spike=spike)$read_count
  meth_spec <- list("mean" = base::mean(methreads/totalreads),
                    "median" = stats::median(methreads/totalreads))
  return(meth_spec)
}

#' create a tiled representation of a genome from the BAM/CRAM file
#' 
#' This function replaces a bedtools call: 
#' bedtools intersect -wao -a fragments.bed -b hg38_300bp_windows.bed > data.bed
#' 
#' The idea is to skip the BED creation step for most runs, and just do it once.
#' In order to count reads in bins, we need bins.  
#' In order to have bins, we need to know how long the chromosomes are. 
#' In order to have a BAM or CRAM file, we need to have those same lengths.
#' This function takes advantage of all of the above to create binned ranges.
#' Note that a very recent branch of Rsamtools is required for CRAM file bins.
#' 
#' @param x       a BAM or CRAM filename (or a BamFile object)
#' @param width   the width of the bins to tile (default is 300) 
#' @param param   optional ScanBamParam (whence we attempt to extract `which`)
#' @param which   an optional GRanges restricting the bins to certain locations 
#' @param ...     additional arguments to pass on to seqinfo_from_header
#' 
#' @return        a GRangesList with y-base-pair-wide bins tiled across it
#'
#' @examples
#' library(Rsamtools) 
#' fl <- system.file("extdata", "ex1.bam", package="Rsamtools", mustWork=TRUE)
#' bam_to_bins(fl) 
#'
#' @seealso seqinfo_from_header
#' 
#' @import IRanges
#' @import GenomicRanges
#' @import GenomeInfoDb
#' @import Rsamtools 
#' 
#' @export
bam_to_bins <- function(x, width=300, param=NULL, which=IRangesList(), ...) { 
 
  gr <- seqinfo_from_header(x, ret="gr", ...)
  if (!is.null(param)) which <- bamWhich(param) 
  if (length(which) > 0) gr <- subsetByOverlaps(gr, which)
  .renameBins(sort(unlist(tile(gr, width=width))), width=width)

}


# helper fn
.renameBins <- function(gr, width) {
  # sls <- seqlevelsStyle(gr)[1]
  # seqlevelsStyle(gr) <- "UCSC"
  names(gr) <- paste(paste0(width, "bp"), paste0("bin", seq_along(gr)), sep="_")
  # seqlevelsStyle(gr) <- sls
  return(gr)
}

#' tabulate methylation specificity for multiple spike-in BAM/CRAM files
#' 
#' Methylation specificity is here defined as methylated_spike_covg/spike_covg
#' 
#' @param files   a vector of BAM/CRAM file names
#' @param spike   a spike-in database 
#' @param sep     the separator for spike-in contig names ("_") 
#' 
#' @return        a matrix with columns "mean" and "median"
#'
#' @examples
#' data(spike) 
#' library(GenomicRanges)
#' sb <- system.file("extdata", "example.spike.bam", package="spiky", 
#'                   mustWork=TRUE)
#' scan_methylation_specificity(sb, spike=spike)
#' 
#' @export
scan_methylation_specificity <- function(files, spike, sep="_") {

  stopifnot(all(file.exists(files)))
  ssc <- lapply(files, spike_counts, spike=spike, dump_idx=TRUE)
  if (is.null(names(ssc))) {
    names(ssc) <- gsub("\\.(bam|cram)", "", basename(files))
  }
  t(vapply(ssc, .tally, numeric(2)))

}


# helper fn
.tally <- function(sc) { 

  methylated <- rowsum(subset(sc, methylated==1)$mapped, 
                       .rfg(subset(sc, methylated==1)))
  total <- rowsum(sc$mapped, .rfg(sc))
  c(mean=mean(methylated/total), median=median(methylated/total))

}


# helper fn (row frag grp)
.rfg <- function(x) {

  gsub("[a-z]", "", tolower(sapply(strsplit(rownames(x), "\\-"), `[`, 1)))

}

#' Scan spikes BEDPE
#'
#'
#' @param bedpe       the BEDPE file path, or output from read_bedpe()
#' @param spike       information about the spikes (default: load `spike`)
#' @param how       how to summarize the per-spike coverage (max)
#' @return          a GRanges with coverage
#'
#' @examples
#' data(spike, package="spiky")
#' fl <- system.file("extdata", "example_spike_bedpe.bed.gz", package="spiky",mustWork=TRUE)
#' scan_spike_bedpe(fl,spike=spike) # will warn user about spike contigs
#'
#' @export
scan_spike_bedpe <- function(bedpe, spike, how="max"){

  if (!is(spike, "DFrame")) stop("Please provide a spike database")
  if (!is(bedpe, "Pairs")) {bedpe <- read_bedpe(bedpe)}
  s <- convertPairedGRtoGR(bedpe)
  seqlevels(s) <- sub("_mod_meth", "", seqlevels(s))
  seqlevels(s) <- sub("_meth", "", seqlevels(s))
  s_covg <- coverage(s)
  s_depth <- get_spike_depth(s_covg,spike=spike,how=how)
  names(s_depth) <- get_base_name(seqnames(s_depth))
  return(as(s_depth,"GRanges"))

}

#' reshape `scan_spiked_bam` results into data.frames for model_glm_pmol
#' @param   spike_gr    GRanges of spike contigs (e.g. output object from scan_spiked_bam, scan_spike_contigs, or scan_spike_bedpe)
#' @param spike spike database (as from data(spike, package="spiky"))
#' @param meth  only keep methylated spike reads? (TRUE; if FALSE, sum both)
#' @param ID    an identifier for this sample, if running several (autogenerate)
#'
#' @return      a data.frame with columns 'frag_grp', 'id', and 'read_count'
#'
#' @seealso     scan_spiked_bam
#'
#' @examples
#' data(spike, package="spiky")
#' data(spike_res, package="spiky")
#' subsetted <- covg_to_df(spike_res, spike=spike, meth=TRUE)
#' summed <- covg_to_df(spike_res, spike=spike, meth=FALSE)
#' round((summed$read_count - subsetted$read_count) / summed$read_count, 3)
#'
#' @export
covg_to_df <- function(spike_gr, spike, meth=TRUE, ID=NULL) {

  if (is.null(ID)) { # mostly to keep from going insane in testing
    ID <- paste(toupper(sample(c(letters, 0:9), 4)), collapse="")
  }

  if (!"methylated" %in% names(mcols(spike_gr))) {
    if (is.null(spike)) {
      message("You need to provide a `spike` database to proceed.")
      message("Consider using the defaults, i.e. `data(spike, package='spiky')`.")
      stop("Cannot assign methylated/unmethylated status to spike contigs.")
    }
    spike_gr$methylated <- spike[names(spike_gr), "methylated"]
  }
  spike_gr$stub <- vapply(strsplit(names(spike_gr), "\\-"), `[`, character(1), 1)
  spike_gr <- as.data.frame(mcols(spike_gr[, c("stub","coverage","methylated")]))
  spike_gr$id <- ID

  if (meth) {
    spike_gr <- subset(spike_gr, spike_gr$methylated == 1)
    spike_gr <- spike_gr[, c("stub", "id", "coverage")]
    names(spike_gr) <- c("frag_grp", "id", "read_count")
  } else {
    spike_gr <- data.frame(id=ID,
                         read_count=tapply(spike_gr$coverage, spike_gr$stub, sum))
    spike_gr$frag_grp <- rownames(spike_gr)
    spike_gr <- spike_gr[, c("frag_grp", "id", "read_count")]
  }

  return(spike_gr[order(spike_gr$frag_grp), c("frag_grp", "id", "read_count")])

}

#' for BAM/CRAM files with renamed contigs, we need to rename `spike` rows
#'
#' This function does that.
#'
#' @param x         a BAM/CRAM file, hopefully with an index
#' @param spike     a DataFrame where spike$sequence is a DNAStringSet
#'
#' @return          a DataFrame with renamed contigs (rows)
#'
#' @seealso generate_spike_fasta
#'
#' @import Rsamtools
#' @import Biostrings
#'
#' @export
rename_spikes <- function(x, spike) {

  # FIXME: code duplication from previous generate_spike_fasta
  cram <- x
  if (!is(cram, "BamFile")) cram <- BamFile(cram)
  hdr <- scanBamHeader(cram)
  cram_contigs <- names(hdr$targets)
  true_contigs <- get_base_name(cram_contigs)
  names(true_contigs) <- cram_contigs

  orphans <- names(which(!true_contigs %in% rownames(spike)))
  if (!is.null(orphans)) {
    message("Orphan contigs found:")
    message(paste(orphans, collapse=", "))
    message("Only a subset of sequences will be represented in output!")
  }

  true_contigs <-intersect(true_contigs, rownames(spike))
  newspikes <- spike[true_contigs, ]
  rownames(newspikes) <- cram_contigs
  names(newspikes$sequence) <- cram_contigs
  return(newspikes)

}

#' decode fragment identifiers for spike-in standards
#' 
#' given a vector of fragment identifiers like `160_2_35` or `80b_1C_35G-2`,
#' encoded typically as lengthInBp_numberOfCpGs_GCpercent, and optionally a 
#' database of spike-in sequences corresponding to those fragments, add those
#' columns to the source data (along with, if present in the database, other 
#' metadata such as standard concentrations, GC fraction, etc.) and return i
#' an updated DataFrame. 
#' 
#' @param x         data.frame with a column of spike information (see above)
#' @param frag_grp  column name for the spike contig information (`frag_grp`)
#' @param spike     optional database of spike-in properties (none)
#' 
#' @return          the data.frame x, augmented with metadata columns
#' 
#' @examples 
#' data(spike_cram_counts) 
#' data(spike, package="spiky") 
#' spike <- subset(spike, methylated == 1)
#' add_frag_info(spike_cram_counts, spike=spike)
#'
#' @export 
add_frag_info <- function(x, frag_grp="frag_grp", spike=NULL) { 

  fg <- gsub("[a-z]+", "", x[, frag_grp], ignore.case=TRUE)
  if (!is.null(spike) && (!frag_grp %in% names(spike))) {
    spike[, frag_grp] <- .tidyfg(rownames(spike))
  }
  frag_grp_parse <- data.frame(do.call(rbind, strsplit(fg, "_")))
  names(frag_grp_parse) <- c("fraglen", "CpG", "GC")
  for (i in names(frag_grp_parse)) x[, i] <- as.integer(frag_grp_parse[, i])
  if (!is.null(spike)) { # can override the above
    for (name in setdiff(names(spike), "sequence")) {
      x[, name] <- spike[match(x[, frag_grp], spike[, frag_grp]), name]
    }
  }
  return(x) 

}

#' Binned estimation of picomoles of DNA present in cfMeDIP assays
#'
#' Given the results of model_glm_pmol and predict_pmol, adjust the predictions
#' to reflect picomoles of captured DNA overlapping a given bin in the genome.
#'
#' @param x     results from predict_pmol (a data.frame or GRanges)
#'
#' @return      the same object, but with a column `adjusted_pred_con`
#'
#' @examples
#' data(spike, package="spiky")
#' data(spike_res, package="spiky")
#' data(genomic_res,package="spiky")
#' fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
#' pred <- predict_pmol(fit, genomic_res, ret="df")
#' bin_pmol(pred)
#'
#' @seealso model_glm_pmol
#' @seealso predict_pmol
#'
#' @export
bin_pmol <- function(x) {

  nameses <- names(x)
  if (is(x, "GRanges")) nameses <- names(mcols(x))
  if (!all(c("pred_conc", "read_count") %in% nameses)) {
    stop("Please use output from predict_pmol as input to bin_pmol.")
  }
  x$adjusted_pred_con <- x$pred_conc * x$read_count
  return(x)

}

#' find spike-in seqlevels in an object `x`, where !is.null(seqinfo(x))
#'
#' Find the spike-like contigs in a BAM with both natural and spiked contigs.
#' This started out as glue in some other functions and got refactored out.
#'
#' @param x       something with seqlevels
#' @param spike   a DataFrame with spike-in information
#'
#' @return        indices of which contigs in seqlevels(x) are spike-in contigs
#'
#' @details
#' The indices have an attribute "mappings", which is a character vector
#' such that attr(find_spike_contigs(x), "mappings") == standardized
#' for all contig names in the CRAM/BAM/whatever, and standardized is
#' the rowname in `spike` that corresponds to the original contig name.
#'
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' si <- seqinfo_from_header(sb)
#' data(spike, package="spiky")
#' find_spike_contigs(si, spike=spike)
#'
#' @seealso get_base_name
#' @seealso rename_spike_seqlevels
#'
#' @export
find_spike_contigs <- function(x, spike) {

  orig_contigs <- seqlevels(x)
  new_contigs <- get_base_name(seqlevels(x))
  #new_contigs <- orig_contigs
  res <- which(new_contigs %in% rownames(spike))

  mappings <- new_contigs
  if (length(new_contigs) == length(orig_contigs)) {
    names(mappings) <- orig_contigs
  } else {
    warning("Cannot resolve contig name mismatches with spike database names. Your file may not have any spikes, or the spike database might not contain them.\n")
  }

  attr(res, "mappings") <- mappings[res]
  return(res)

}

#' simple contig kmer comparisons
#'
#' @param km          kmer summary
#' @param normalize   normalize (divide by row sums)? (TRUE)
#'
#' @return            the most common kmers for each contig, across all contigs
#'
#' @examples
#'
#'
#' data(genbank_mito, package="spiky")
#' mtk6 <- kmers(genbank_mito, k=6)
#' rownames(mtk6) <- paste0(rownames(mtk6), "_MT")
#' kmax(mtk6)
#'
#' data(phage, package="spiky")
#' phk6 <- kmers(phage, k=6)
#' kmax(phk6, normalize=FALSE)
#'
#' stopifnot(identical(colnames(phk6), colnames(mtk6)))
#' k6 <- rbind(mtk6, phk6)
#' kmax(k6)
#'
#' @export
kmax <- function(km, normalize=TRUE) {

  if (length(unique(nchar(colnames(km)))) > 1) stop("kmax is for kmer tallies!")
  if (normalize) km <- sweep(km, 1, rowSums(km), `/`)
  km[seq_len(nrow(km)), unique(max.col(km))]

}

#' oligonucleotideFrequency, but less letters and more convenient.
#' 
#' @param  x          BSgenome, DFrame with `sequence` column, or DNAStringSet
#' @param  k          the length of the kmers (default is 6)
#' 
#' @return            a matrix of contigs (rows) by kmer frequencies (columns)
#'
#' @details 
#' The companion `kmax` function finds the maximum frequency kmer for each
#' contig and plots all of them together for comparison purposes. 
#' 
#' @examples
#' 
#' data(genbank_mito, package="spiky") 
#' mtk6 <- kmers(genbank_mito, k=6)
#' kmax(mtk6)
#' 
#' data(phage, package="spiky") 
#' phk6 <- kmers(phage, k=6)
#' kmax(phk6)
#' 
#' @seealso           kmax
#' 
#' @import            Biostrings
#' 
#' @export
kmers <- function(x, k=6) {

  if (is(x, "DNAStringSet")) {
    res <- oligonucleotideFrequency(x, width=k)
    rownames(res) <- names(x)
  } else if (is(x, "DFrame")) { 
    res <- oligonucleotideFrequency(x$sequence, width=k)
    rownames(res) <- rownames(x)
  } else if (is(x, "BSgenome")) {
    chrs <- names(x)
    names(chrs) <- chrs
    res <- t(sapply(chrs, 
                    function(chr) oligonucleotideFrequency(x[[chr]], k)))
  } else { 
    stop("Don't know how to get kmers for an object of type ", class(x))
  } 

  attr(res, "what") <- "kmers"
  return(res)

}

#' use the index of a spiked BAM/CRAM file for spike contig coverage 
#' 
#' It dawned on me one day that we don't even have to bother reading the file
#' if we have an index for a spiked BAM/CRAM result, since any fragments that 
#' map properly to the spike contigs are generated from synthetic templates.
#' This function takes an index and a spike database (usually a DataFrame) as
#' inputs and provides a rough coverage estimate over "rehabilitated" contig
#' names (i.e., canonicalized contigs mapping to the database) as its output. 
#' 
#' @param     bam             the BAM or CRAM file (MUST HAVE AN INDEX) 
#' @param     spike           a data.frame, DataFrame, or similar with spikes
#' @param     sep             separator character in contig names ("_")
#' @param     ref             reference name for spike genome ("spike")
#' @param     verbose         be verbose? (FALSE) 
#' @param     dump_idx        dump the renamed idxstats to aggregate? (FALSE)
#' 
#' @examples
#' data(spike, package="spiky")
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE) 
#' spike_counts(sb, spike=spike)
#' 
#' @details
#' The argument `spike` has no default since we are attempting to refactor the
#' spike-in databases into their own data packages and allow more general use.
#' 
#' @return                    a GRanges of spike-in contig read counts
#'
#' @import GenomicRanges
#' @import Rsamtools
#' 
#' @export
spike_counts <- function(bam, spike, sep="_", ref="spike", verbose=FALSE, dump_idx=FALSE) {

  if (!is(bam, "BamFile")) bam <- BamFile(bam)
  if (!file.exists(bam$index)) {
    message("You need a valid index for ", bam$path, " to use this function.")
    message("(Please do NOT use Rsamtools::indexBam() if it is a CRAM file!)")
    stop(bam$index, " was not found; aborting.")
  }
  if (!all(c("sequence", "methylated", "CpGs") %in% names(spike)) |
      !all(grepl(sep, rownames(spike)))) {
    stop("Your spike-in database is missing either columns or names; aborting.")
  }

  if (verbose) message("Scanning ", bam$index, "...")
  isb <- idxstatsBam(bam) 
  sbn <- get_base_name(isb[, "seqnames"])
  spike_contigs <- sbn[which(sbn %in% rownames(spike))]
  isb <- subset(isb, seqnames %in% names(spike_contigs))
  isb$seqnames <- names(spike_contigs)
  rownames(isb) <- spike_contigs
  isb$methylated <- spike[rownames(isb), "methylated"] 
  isb$id <- gsub("\\.(bam|cram)", "", basename(bam$path))
  if (verbose) message(length(spike_contigs), " spike-in contigs identified.")

  # used by scan_spike_counts
  if (dump_idx) {
    return(isb)
  } else { 
    # prep seqinfo
    isbsi <- as(isb, "Seqinfo") 
    seqlengths(isbsi) <- isb$seqlength
    genome(isbsi) <- ref

    # prep covg
    isb$start <- 1
    isb$end <- isb$seqlength
    isb$unmapped <- NULL # irrelevant
    names(isb) <- sub("mapped", "coverage", names(isb))
    covg <- as(isb[, c("seqnames", "start", "end", "coverage", "methylated")], 
               "GRanges")
    seqinfo(covg) <- isbsi[seqlevels(covg)]

    # done
    return(covg)
  }

}

#' pretty much what it says: scan spike contigs from a BAM or CRAM file
#'
#' default workflow is
#'
#' 1. scan spike contigs and count fragments per contig or per bin.
#' 2. fit the appropriate model for adjusting genomic contigs based on spikes.
#' 3. scan and adjust binned fragment tallies along genomic contigs per above.
#'
#' scan_spike_contigs implements step 1.

#'
#' @param bam       the BAM or CRAM filename, or a vector of such filenames
#' @param spike     the spike-in reference database (e.g. data(spike))
#' @param how       how to summarize the per-spike coverage (max)
#' @param param     a ScanBamParam object, or NULL (will default to MAPQ=20 etc)
#' @param ...       additional arguments to pass to scanBamFlag()
#' @param mc.cores  Number of cores to run on (default 16)
#'
#' @details
#' If multiple BAM or CRAM filenames are provided, all indices will be
#' checked before attempting to run through any of the files.
#'
#' @return          a CompressedGRangesList with bin- and spike-level coverage
#'
#' @examples
#' library(GenomicRanges)
#' data(spike, package="spiky")
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE) # switch to a CRAM
#' res <- scan_spike_contigs(sb, spike=spike) # use default ScanBamParam
#' summary(res)
#'
#' @seealso         Rsamtools::ScanBamParam
#'
#' @import          GenomicAlignments
#' @import          GenomeInfoDb
#' @import          Rsamtools
#'
#' @export
scan_spike_contigs <- function(bam, spike, how="max", param=NULL, mc.cores=16,...) {

  # can be smoother but:
  if (length(bam) > 1) {
    indices <- sub("bam$", "bam.bai", bam)
    indices <- sub("cram$", "cram.crai", bam)
    if (!all(file.exists(indices))) {
      missed <- indices[!file.exists(indices)]
      stop("Missing index files: ", paste(missed, collapse=", "))
    } else {
      if (is.null(names(bam))) names(bam) <- bam
      return(lapply(bam, scan_spike_contigs, spike=spike))
    }
  }

  # scan the BAM (or CRAM if supported) to determine which reads to import
  si <- seqinfo_from_header(bam)
  spikes <- find_spike_contigs(si, spike=spike)
  if (length(spikes) > 0) {
    mappings <- attr(spikes, "mappings")
    attr(spikes, "mappings") <- NULL
    genome(si)[spikes] <- "spike"
  }

  # create appropriate filters for coverage tabulation if param=NULL
  if (is.null(param)) {
    fl <- scanBamFlag(isDuplicate=FALSE,
                      isPaired=TRUE, ...)
    param <- ScanBamParam(flag=fl)
    bamMapqFilter(param) <- 20
  }

  orig_spike_contigs <- subset(seqlevels(si), genome(si) == "spike")
  if (length(orig_spike_contigs) == 0) {
    # empty coverage list
    s <- as(S4Vectors::SimpleList(), "SimpleRleList")
  } else {
    # rationalize the contigs but don't override user supplied bamWhich
    gr <- as(sortSeqlevels(si[orig_spike_contigs]), "GRanges") # kludgey
    if (length(bamWhich(param)) == 0) bamWhich(param) <- gr
    s <- GenomicAlignments::coverage(BamFile(bam), param=param)
  }
  s_depth <- get_spike_depth(s,spike=spike,how=how)
  return(as(s_depth,"GRanges"))
}

#' parse out the forward and reverse UMIs and contig for a BED/BAM
#'
#' @param   UMI   a vector of UMIs
#' @param   pos   optional vector of positions (else all are set to 1) 
#' @param   seqs  optional vector of read sequences (else widths default to 96)
#' 
#' @return  a GRanges
#' 
#' @import Biostrings
#' @import GenomicRanges
#' 
#' @export 
parse_spike_UMI <- function(UMI, pos=NULL, seqs=NULL) { 
  nch <- nchar(UMI)
  UMIs <- strsplit(substr(UMI, nch - 10, nch), "_")
  x <- data.frame(UMI1 = DNAStringSet(vapply(UMIs, `[`, character(1), 1)),
                  UMI2 = DNAStringSet(vapply(UMIs, `[`, character(1), 2)),
                  chrom = substr(UMI, 1, nch - 11))

  x$chromStart <- 1
  if (!is.null(pos)) x$chromStart <- pos
  widths <- rep(96, nrow(x))
  if (!is.null(seqs)) x$seq <- DNAStringSet(seqs)
  if (!is.null(seqs)) widths <- width(x$seq)
  x$chromEnd <- x$chromStart + widths - 1 
  x$name <- UMI 
  x$score <- 1 # subject 6547 

  return(as(x, "GRanges"))
} 

#' Build a Bayesian additive model from spike-ins to correct bias in *-seq
#'
#' @param   x     data with assorted feature information (GCfrac, CpGs, etc)
#' @param   conc  concentration for each spike (must be provided!)
#' @param   fm    model formula (conc ~ read_count + fraglen + GCfrac + CpGs_3)
#' @param   ...   other arguments to pass to `bamlss`
#'
#' @return        the model fit for the data
#'
#' @examples
#'
#' library(bamlss)
#' data(spike_cram_counts,package="spiky")
#' data(spike,package="spiky")
#' scc <- add_frag_info(spike_cram_counts, spike=spike)
#' scc$conc <- scc$conc * 0.9 # adjust for dilution
#' scc$CpGs_3 <- scc$CpGs ^ (1/3)
#' fit0 <- model_bam_standards(scc,
#'                             fm=conc ~ read_count + fraglen)
#' fit1 <- model_bam_standards(scc,
#'                             fm=conc ~ read_count + fraglen + GCfrac + CpGs_3)
#' DIC(fit0, fit1)
#'
#' @import        bamlss
#'
#' @export
model_bam_standards <- function(x, conc=NULL, fm=NULL, ...) {

  if (is.null(fm)) fm <- conc ~ read_count + fraglen + GCfrac + CpGs_3
  if (is.null(x$conc) & !is.null(conc)) x$conc <- conc

  cols <- strsplit(as.character(fm)[3], " + ", fixed=TRUE)[[1]]
  if (!all(cols %in% names(x))) {
    stop("Input data must have columns ", paste(cols, collapse=", "))
  }

  # Gaussian model by default -- can alter with ... params
  bamlss(fm, data=x, ...)

}

#' QC, QA, and processing for a new spike database
#'
#' Sequence feature verification: never trust anyone, least of all yourself.
#'
#' @param fasta         fasta file (or GRanges or DataFrame) w/spike sequences
#' @param methylated    whether CpGs in each are methylated (0 or 1, default 0)
#' @param ...           additional arguments, e.g. kernels (currently unused)
#'
#' @return              a DataFrame suitable for downstream processing
#'
#' @details
#' GCfrac is the GC content of spikes as a proportion instead of a percent.
#' OECpG is (observed/expected) CpGs (expectation is 25% of GC dinucleotides).
#'
#' @examples
#'
#' data(spike)
#' spikes <- system.file("extdata", "spikes.fa", package="spiky", mustWork=TRUE)
#' spikemeth <- spike$methylated
#' process_spikes(spikes, spikemeth)
#'
#' data(phage)
#' phages <- system.file("extdata", "phages.fa", package="spiky", mustWork=TRUE)
#' identical(process_spikes(phage), phage)
#' identical(phage, process_spikes(phage))
#'
#' data(genbank_mito)
#' (mt <- process_spikes(genbank_mito)) # see also genbank_mito.R
#' gb_mito <- system.file("extdata", "genbank_mito.R", package="spiky")
#'
#'
#' @import Biostrings
#' @import S4Vectors
#'
#' @seealso kmers
#'
#' @export
process_spikes <- function(fasta, methylated=0, ...) {

  if (is.character(fasta) && file.exists(fasta)) {
    spikes <- DataFrame(sequence=readDNAStringSet(fasta), methylated=methylated)
  } else if (is(fasta, "GRanges") && "sequence" %in% names(mcols(fasta))) {
    spikes <- fasta
    if (!"methylated" %in% names(mcols(fasta))) spikes$methylated <- methylated
  } else if (is(fasta, "DataFrame") && "sequence" %in% names(fasta)) {
    spikes <- fasta
    if (!"methylated" %in% names(fasta)) spikes$methylated <- methylated
  } else {
    stop(fasta, " does not exist or is not in a suitable format to proceed.")
  }

  spikes$CpGs <- dinucleotideFrequency(spikes$sequence)[, "CG"]
  GC <- apply(alphabetFrequency(spikes$sequence)[, c("C", "G")], 1, sum)
  spikes$GCfrac <- round(GC / width(spikes$sequence), 2)
  spikes$OECpG <- round(spikes$CpGs / (GC/8), 2) # CpG expected 25% of GC dinuc
  return(spikes)

}

#' scan genomic contigs in a BAM/CRAM file
#'
#' The default workflow for spiky is roughly as follows:
#'
#' 1. Identify and quantify the spike-in contigs in an experiment.
#' 2. Fit a model for sequence-based abundance artifacts using the spike-ins.
#' 3. Quantify raw fragment abundance on genomic contigs, and adjust per step 2.
#'
#' scan_genomic_contigs addresses the first half of step 3. The assumption is
#' that anything which isn't a spike contig, is a genomic contig.  This isn't
#' necessarily true, so the user can also supply a ScanBamParam object for the
#' `param` argument and restrict scanning to whatever contigs they wish, which
#' also allows for non-default MAPQ, pairing, and quality filters.
#'
#' @param bam       the BAM or CRAM filename, or a vector of them
#' @param spike     the spike-in reference database (e.g. data(spike))
#' @param param     a ScanBamParam object specifying which reads to count (NULL)
#' @param bin       Bin reads? (TRUE)
#' @param binwidth  width of the bins for chromosomal tiling (300)
#' @param bins      a pre-tiled GRanges for binning coverage (NULL)
#' @param standard  restrict non-spike contigs to "standard" chromosomes? (TRUE)
#' @param genome    Name of genome (default hg38)
#' @param ...       additional arguments to pass to scanBamFlag()
#'
#' @return          a CompressedGRangesList with bin- and spike-level coverage
#'
#' @details
#' If multiple BAM or CRAM filenames are provided, all indices will be
#' checked before attempting to run through any of the files.
#'
#' @examples
#'
#' library(Rsamtools)
#' data(spike, package="spiky")
#'
#' fl <- system.file("extdata", "ex1.bam", package="Rsamtools",
#'                   mustWork=TRUE)
#' scan_genomic_contigs(fl, spike=spike,standard=FALSE) # will warn user about spike contigs
#'
#' sb <- system.file("extdata", "example_chr21.bam", package="spiky",
#'                   mustWork=TRUE)
#' scan_genomic_contigs(sb, spike=spike) # will warn user about genomic contigs
#'
#' @seealso         Rsamtools::ScanBamParam
#' @import          GenomicAlignments
#' @import          Rsamtools
#'
#' @export
scan_genomic_contigs <- function(bam, spike, param=NULL,bin=TRUE, binwidth=300L, bins=NULL, standard=TRUE,genome="hg38", ...) {

  # can be smoother but:
  if (length(bam) > 1) {
    indices <- sub("bam$", "bam.bai", bam)
    indices <- sub("cram$", "cram.crai", bam)
    if (!all(file.exists(indices))) {
      missed <- indices[!file.exists(indices)]
      stop("Missing index files: ", paste(missed, collapse=", "))
    } else {
      if (is.null(names(bam))) names(bam) <- bam
      return(lapply(bam, scan_genomic_contigs, spike=spike))
    }
  }

  # scan the BAM (or CRAM if supported) to determine which reads to import
  si <- seqinfo_from_header(bam)
  if (standard) seqlevels(si) <- seqlevels(keepStandardChromosomes(si))
  mappings <- attr(find_spike_contigs(si, spike=spike), "mapping")
  spike_contigs <- names(mappings)
  genomic_contigs <- seqlevels(si)
  if (length(spike_contigs) > 0) {
    genomic_contigs <- setdiff(genomic_contigs, spike_contigs)
  }
  if (length(genomic_contigs) == 0) {
    stop(bam, " doesn't appear to have any genomic contigs.")
  }

  bf <- BamFile(bam)
  if (is.null(param)) {
    fl <- scanBamFlag(isDuplicate=FALSE,
                      isPaired=TRUE, ...)
    param <- ScanBamParam(flag=fl)
    bamMapqFilter(param) <- 20
  }

  # rationalize the contigs (but do not replace user-supplied Ranges)
  gr <- as(sortSeqlevels(si[genomic_contigs]), "GRanges") # kludgey
  if (length(bamWhich(param)) == 0) bamWhich(param) <- gr
  # assess coverage on these contigs (bin later)
  g_covg <- GenomicAlignments::coverage(BamFile(bam), param=param)
  genome(gr) <- genome
  if (is.null(bins)) bins <- tile_bins(gr=gr, binwidth=binwidth)

  if (bin){
    # genomic coverage is averaged across each bin
    if (length(bins) > 0) {
      message("Binning genomic coverage...")
      binned <- get_binned_coverage(bins, g_covg)
      message("Done.")
    } else {
      message("Empty bins provided, skipping genomic binning.")
      binned <- bins
    }
  } else {binned <- bins}

  genomic_gr <- as(binned,"GRanges")
  names(genomic_gr) <- seqnames(genomic_gr)
  return(genomic_gr)
}

#' spike-in counts for two samples, as a wide data.frame
#'
#' A data.frame with spike-in results from control samples in the manuscript.
#' This maps 1:1 onto `spike_read_counts` using reshape2::melt.
#'
#' @importFrom utils data
#' @usage data(dedup)
#' @format A data.frame object with
#' \describe{
#'   \item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
#'   \item{read_count_6547}{read coverage for this spike in sample 6547}
#'   \item{read_count_6548}{read coverage for this spike in sample 6548}
#' }
#' @source This data was created using inst/script/loadDedup.R
"dedup"


#' various mitochondrial genomes sometimes used as endogenous spike-ins
#'
#' A DataFrame with species, genome, accession, and sequence for GenBank
#' mitochondrial genome depositions. No concentration provided; add if needed.
#'
#' @usage data(genbank_mito)
#' @format A DataFrame object with
#' \describe{
#'   \item{species}{the species whence the record came, as a character string}
#'   \item{genome}{the genome assembly whence the mtDNA, as a character string}
#'   \item{accession}{the genbank accession, as a character string}
#'   \item{sequence}{genome sequence, as a DNAStringSet}
#' }
#' @source www.ncbi.nlm.nih.gov/genbank/
"genbank_mito"


#' lambda and phiX phage sequences, sometimes used as spike-ins
#'
#' A DataFrame with sequence, methylated, CpGs, GCfrac, and OECpG for phages
#'
#' @usage data(phage)
#' @format A DataFrame object with
#' \describe{
#'   \item{sequence}{genome sequence, as a DNAStringSet}
#'   \item{methylated}{whether CpGs are methylated, as an integer}
#'   \item{CpGs}{the number of CpGs in the phage genome, as an integer}
#'   \item{GCfrac}{the GC fraction of the phage genome, as a numeric}
#'   \item{OECpG}{the observed / expected CpG fraction, as a numeric}
#' }
#' @source www.ncbi.nlm.nih.gov/genbank/
"phage"


#' spike-in contig properties for Sam's cfMeDIP spikes
#'
#' A DataFrame with sequence, concentration, and other properties of Sam's
#' synthetic cfMeDIP spike-in controls. The row names redudantly encode some
#' of these properties, such as the number of CpGs in the spike-in sequence.
#'
#' @usage data(spike)
#' @format A DataFrame object with
#' \describe{
#'   \item{sequence}{contig sequence, as a DNAStringSet}
#'   \item{methylated}{are the CpGs in this spike-in methylated? 0 or 1}
#'   \item{CpGs}{number of CpG dinucleotides in the spike, from 1 to 16}
#'   \item{fmol}{femtomolar concentration of the spike-in for standard mix}
#'   \item{molmass}{molar mass of spike-in sequence}
#' }
#' @source https://doi.org/10.1101/2021.02.12.430289
"spike"


#' spike-in counts, as a long data.frame
#'
#' A data.frame with spike-in results from control samples in the manuscript.
#' This maps 1:1 onto `dedup` using reshape2::melt.
#'
#' @usage data(spike_read_counts)
#' @format A data.frame object with
#' \describe{
#'   \item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
#'   \item{id}{subject from whom cfMeDIP spike reads (column 3) were counted}
#'   \item{read_count}{read coverage for this spike in this subject (column 2)}
#' }
#' @source This data was created using inst/script/loadDedup.R
"spike_read_counts"

#' spike-in counts, as a long data.frame
#'
#' A data.frame with spike-in results from CRAM files (generated from scan_spike_counts(CRAMs, spike=spike))
#'
#' @usage data(spike_cram_counts)
#' @format A data.frame object with
#' \describe{
#'   \item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
#'   \item{id}{subject from whom cfMeDIP spike reads (column 3) were counted}
#'   \item{read_count}{read coverage for this spike in this subject (column 2)}
#' }
#' @source Generated from scan_spike_counts(CRAMs, spike=spike) using example CRAMs
#' containing spike contigs
"spike_cram_counts"


#' scan_spiked_bam results from a merged cfMeDIP CRAM file (chr22 and spikes)
#'
#' A CompressedGRangesList object with `genomic` (chr22) and `spikes` coverage,
#' binned every 300bp for the genomic contigs then averaged across the bin, and
#' summarized for each spike contig as (the default) `max` coverage. (In other
#' words, the default output of scan_spiked_bam, restricted to a small enough
#' set of genomic regions to be practical for examples.) This represents what
#' most users will want to generate from their own merged BAMs or CRAMs, and
#' is used repeatedly in downstream examples throughout the package.
#'
#' @usage data(ssb_res)
#' @format A CompressedGRangesList of coverage results, containing
#' \describe{
#'   \item{genomic}{a GRanges with one metadata column, `coverage`}
#'   \item{spikes}{a GRanges with one metadata column, `coverage`}
#' }
#' @source Generated using scan_spiked_bam on an example bam
#' containing chr22 and spike contigs.
"ssb_res"

#' A Granges object with genomic coverage from chr21q22,
#' binned every 300bp for the genomic contigs then averaged across the bin. (In other
#' words, the default output of scan_genomic_contigs or scan_genomic_bedpe,
#' restricted to a small enough set of genomic regions to be practical for examples.)
#' This represents what most users will want to generate from their own genomic BAMs or BEDPEs, and
#' is used repeatedly in downstream examples throughout the package.
#'
#' @usage data(genomic_res)
#' @format A GRanges of coverage results with one metadata column, `coverage`
#' @source Generated using scan_genomic_bedpe or scan_genomic_contigs on an example bedpe or bam
#' containing chr21q22 contigs.
"genomic_res"

#' A Granges object with spike-in sequence coverage, and
#' summarized for each spike contig as (the default) `max` coverage. (In other
#' words, the default output of scan_spike_contigs or scan_spike_bedpe)
#' This represents what most users will want to generate from their own spike-in BAMs or BEDPEs, and
#' is used repeatedly in downstream examples throughout the package.
#'
#' @usage data(spike_res)
#' @format A GRanges of coverage results with one metadata column, `coverage`
#' @source Generated using scan_spike_bedpe or scan_spike_contigs on an example bedpe or bam
#' containing spike contigs.
"spike_res"

#' a test GRanges with UMI'ed genomic sequences used as controls
#'
#' Sources and overlap widths of various read sequences in a test CRAM.
#'
#' @usage data(testGR)
#' @format A GRanges object with an mcols() DataFrame containing
#' \describe{
#'   \item{UMI1}{the unique molecular identifier on the forward read}
#'   \item{UMI2}{the unique molecular identifier on the reverse read}
#'   \item{seq}{the sequence of the fragment}
#'   \item{name}{the name of the fragment}
#'   \item{score}{whether the fragment passes filters (always 1)}
#' }
#' @source Generated using inst/script/loadTest.R
"testGR"

#' refactored out of rename_spikes and rename_spike_seqlevels
#'
#' A common task between generate_spike_fasta, rename_spikes, and
#' rename_spike_seqlevels is to determine what the largest common subset of
#' characters between existing contig names and stored standardized contigs
#' might be. This function eases that task.
#'
#' @param     contig_names    the names of contigs
#' @param     sep             separator character in contig names ("_")
#'
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' bh <- scanBamHeader(BamFile(sb))
#' orig_contigs <- names(bh$targets)
#' get_base_name(orig_contigs)
#'
#' @return                    a vector of elements 1:3 from each contig name
#'
#' @export
get_base_name <- function(contig_names, sep="_") {

  elts <- seq_len(3)
  contig_names <- as.character(contig_names)
  res <- vapply(lapply(strsplit(contig_names, sep), `[`, elts),
                paste, character(1), collapse=sep)
  names(res) <- contig_names

  keep <- grep("NA", res, invert=TRUE)
  res[keep]
}

#' get the (max, median, or mean) coverage for spike-in contigs from a BAM/CRAM
#'
#' @param covg      the coverage RleList
#' @param spike_gr  the spike-in GRanges (default: figure out from seqinfo)
#' @param spike     information about the spikes (default: load `spike`)
#' @param how       how to summarize the per-spike coverage (max)
#'
#' @return          a GRanges with summarized coverage and features for each
#'
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' data(spike, package="spiky")
#' si <- seqinfo_from_header(sb)
#' genome(si) <- "spike"
#' mgr <- get_merged_gr(si,spike=spike)
#'
#' fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
#' bp <- ScanBamParam(flag=fl)
#' bamMapqFilter(bp) <- 20
#'
#' covg <- get_spiked_coverage(sb, bp=bp, gr=mgr)
#' get_spike_depth(covg, spike_gr=mgr, spike=spike)
#'
#' @export
get_spike_depth <- function(covg, spike_gr=NULL, spike=NULL,
                            how=c("max", "mean")) {

  how <- match.fun(match.arg(how))
  if (!is(spike, "DFrame")) stop("Please provide a spike database")
  cols <- colnames(spike)[-1]

  # as with scan_spike_contigs()
  if (is.null(spike_gr)) {
    si <- seqinfo(covg)
    seqlengths(si) <- sapply(covg, length)
    new_contigs <- attr(find_spike_contigs(covg, spike), "mapping")
    spike_gr <- as(si, "GRanges")
    names(spike_gr) <- new_contigs[names(spike_gr)]
  }
  canon <- names(spike_gr)

  message("Summarizing spike-in counts...", appendLF=FALSE)
  spike_depth <- vapply(covg[seqlevels(spike_gr)], how, numeric(1), na.rm=TRUE)
  for (nm in cols) mcols(spike_gr)[[nm]] <- spike[canon, nm]
  spike_gr$coverage <- spike_depth[as.character(seqnames(spike_gr))]
  message("Done.")

  return(spike_gr)

}

#' Convert Pairs to GRanges
#'
#' @param pairs          the Pairs object
#'
#' @return            a GRanges
#'
#' @export
convertPairedGRtoGR <- function(pairs){
  ir <- IRanges(start=start(first(pairs)),end=end(second(pairs)))
  gr <- GRanges(seqnames=seqnames(first(pairs)),ranges=ir)
  return(gr)
}

#' for CRAM files, a FASTA reference is required to decode; this builds that
#'
#' A FASTA reference is *not* always needed, so long as .crai indices
#' are available for all contigs in the CRAM. See `spike_counts` for a fast
#' and convenient alternative that extracts spike coverage from index stats.
#' However, spike_counts has its own issues, and it's better to use fragments.
#'
#' If the contigs in a CRAM have even slightly different names from those in
#' the reference, decoding will fail.  In some cases there are multiple names
#' for a given contig (which raises the question of whether to condense them),
#' and thus the same reference sequence decodes multiple contig names.
#'
#' This function generates an appropriate spike reference for a BAM or CRAM,
#' using BAM/CRAM headers to figure out which references are used for which.
#'
#' At the moment, CRAM support in Rsamtools only exists in the GitHub branch:
#'
#' BiocManager::install("Bioconductor/Rsamtools@cram")
#'
#' Using other versions of Rsamtools will yield an error on CRAM files.
#'
#' Note that for merged genomic + spike reference BAMs/CRAMs, this function
#' will only attempt to generate a FASTA for the spike contigs, not reference.
#' If your reference contigs are screwed up, talk to your sequencing people,
#' and keep better track of the FASTA reference against which you compress!
#'
#' @param bam       a BAM or CRAM file, hopefully with an index
#' @param spike     the spike contig database (mandatory as of 0.9.99)
#' @param assembly  optional BSgenome or seqinfo with reference contigs (NULL)
#' @param fa        the filename for the resulting FASTA ("spikes.fa")
#'
#' @return          invisibly, a DNAStringSet as exported to `fa`
#'
#' @examples
#'
#' library(GenomicRanges)
#' data(spike, package="spiky")
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' outFasta <- paste(system.file("extdata", package="spiky", mustWork=TRUE),"/spike_contigs.fa",sep="")
#' show(generate_spike_fasta(sb, spike=spike,fa=outFasta))
#'
#' @seealso         rename_contigs
#'
#' @import          Biostrings
#' @import          Rsamtools
#'
#' @export
generate_spike_fasta <- function(bam, spike, assembly=NULL, fa="spike_contigs.fa"){

  if (!is(bam, "BamFile")) bam <- BamFile(bam)
  hdr <- scanBamHeader(bam)
  bam_contigs <- names(hdr$targets)
  if (!is.null(assembly)) {
    bam_contigs <- setdiff(bam_contigs, seqlevels(assembly))
  }

  # are there spikes in it?
  if (length(bam_contigs) == 0) stop("No spike contigs found in ", bam, "!")
  newspikes <- rename_spikes(bam, spike=spike) # this is probably dumb

  contigs <- newspikes[bam_contigs, "sequence"]
  names(contigs) <- bam_contigs
  writeXStringSet(contigs, fa)
  message("Wrote ", fa)
  invisible(contigs)

}


# helper fn
.get_base_name <- function(contig_names, sep="_") {

  vapply(lapply(strsplit(contig_names, sep), `[`, seq_len(3)),
         paste, character(1), collapse=sep)

}

#' tabulate coverage across assembly and spike contig subset in natural order
#'
#' FIXME: this is wicked slow, ask Herve if a faster version exists 
#' 
#' Refactored from scan_spiked_bam, this is a very simple wrapper
#'
#' @param bf  the BamFile object 
#' @param bp  the ScanBamParam object 
#' @param gr  the GRanges with sorted seqlevels
#' 
#' @return    a list of Rles 
#' 
#' @seealso   scan_spiked_bam
#' @seealso   coverage
#' 
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky", 
#'                   mustWork=TRUE)
#' si <- seqinfo_from_header(sb) 
#' genome(si) <- "spike"
#
#' data(spike, package="spiky")
#' mgr <- get_merged_gr(si, spike=spike) # note canonicalized spikes
#'
#' fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
#' bp <- ScanBamParam(flag=fl)
#' bamMapqFilter(bp) <- 20
#' get_spiked_coverage(sb, bp=bp, gr=mgr)
#'
#' @import    GenomicAlignments
#' 
#' @export
get_spiked_coverage <- function(bf, bp, gr) { 
  
  message("Tabulating read pair coverage (may take a while)...", appendLF=FALSE)
  covg <- coverage(bf, param=bp)[seqlevels(gr)]
  message("Done.") 
  return(covg)

}

#' Scan genomic BEDPE
#'
#'
#' @param bedpe       the BEDPE file path, or output from read_bedpe()
#' @param bin       Bin reads? (TRUE)
#' @param binwidth  width of the bins for chromosomal tiling (300)
#' @param bins      a pre-tiled GRanges for binning coverage (NULL)
#' @param standard  restrict non-spike contigs to "standard" chromosomes? (TRUE)
#' @param genome    Name of genome (default hg38)
#'
#' @return          a GRanges with coverage
#'
#' @examples
#'
#' fl <- system.file("extdata", "example_chr21_bedpe.bed.gz", package="spiky",mustWork=TRUE)
#' scan_genomic_bedpe(fl) # will warn user about spike contigs
#'
#'
#' @export
scan_genomic_bedpe <- function(bedpe, bin=TRUE, binwidth=300L, bins=NULL, standard=TRUE,genome="hg38"){
  if (!is(bedpe, "Pairs")) {bedpe <- read_bedpe(bedpe)}
  genomic <- convertPairedGRtoGR(bedpe)

  g_covg <- coverage(genomic)

  g_si <- seqinfo(g_covg)
  seqlengths(g_si) <- sapply(g_covg, length)
  if (standard) seqlevels(g_si) <- seqlevels(keepStandardChromosomes(g_si))
  genome(g_si) <- genome
  g_gr <- as(g_si,"GRanges")
  if (is.null(bins)) bins <- tile_bins(gr=g_gr, binwidth=binwidth)

  if (bin){
    # genomic coverage is averaged across each bin
    if (length(bins) > 0) {
      message("Binning genomic coverage...")
      binned <- get_binned_coverage(bins, g_covg)
      message("Done.")
    } else {
      message("Empty bins provided, skipping genomic binning.")
      binned <- bins
    }
  } else {binned <- bins}

  gr <- as(binned,"GRanges")
  names(gr) <- seqnames(gr)
  return(gr)

}

#' for spike-in contigs in GRanges, match to standardized spike seqlevels
#' 
#' This function is essentially the opposite of rename_spikes, except that 
#' it works well on GRanges/GAlignments from or for merged genome+spike BAMs.
#' If spike contigs are found, it will assign genome='spike' to those, while
#' changing the seqlevels to standardized names that match rownames(spike). 
#' 
#' @param x         something with seqlevels (GRanges, GAlignments, Seqinfo...)
#' @param spike     a DataFrame where spike$sequence is a DNAStringSet (or NULL)
#' 
#' @return          x, but with standardized spike seqlevels and genomes
#' 
#' @seealso         rename_spikes
#' 
#' @import Rsamtools
#' 
#' @export
rename_spike_seqlevels <- function(x, spike=NULL) { 

  has_spike <- find_spike_contigs(x, spike=spike)
  if (length(has_spike) > 0) { 
    genome(x)[has_spike] <- "spike"
    new_contigs <- attr(has_spike, "mappings")
    old_contigs <- seqlevels(x)[has_spike] 
    seqlevels(x)[has_spike] <- new_contigs[old_contigs] 
    genome(x)[has_spike] <- "spike"
  } else { 
    message("No spike-in contigs were found.")
  }
  return(x)

}

#' Bland-Altman plot for cfMeDIP spike standards
#'
#' @param   fit             a model fit, from predict_pmol (?)
#'
#' @return                  a ggplot2 object
#'
#' @import  BlandAltmanLeh
#' @import  ggplot2
#' @import  scales
#'
#' @examples
#'
#' data(spike_res)
#' data(spike, package="spiky")
#' fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
#' ba_plot <- spike_bland_altman_plot(fit)
#'
#' @export
spike_bland_altman_plot <- function(fit) {

  if (!is.data.frame(fit$x)) fit$x <- attr(fit, "data")
  fit$x$pred_conc <- predict(fit, fit$x)

  fit$x$fraglen <- as.factor(paste0(fit$x$fraglen, "bp"))
  BA <- bland.altman.plot(fit$x$conc,
                          fit$resid,
                          conf.int = .95,
                          col.points = fit$x$fraglen,
                          pch = 19,
                          graph.sys = "ggplot2")

  BA <- BA +
    aes(color = fit$x$fraglen,shape = fit$x$fraglen) +
    theme_bw(base_line_size = 0.25) +
    scale_color_manual(values = c("80bp" = "black",
                                  "160bp" = "darkgray",
                                  "320bp" = "navy")) +
    theme(legend.title = element_blank(),
          legend.position = "right",
          legend.text = element_text(size = 14),
          axis.title.y = element_text(size = 20),
          axis.text.x  = element_text(size = 16),
          axis.text.y = element_text(vjust = 0.5, size = 16),
          axis.title.x = element_text(size = 20),
          panel.grid.minor.x = element_blank()) +
    xlab("Mean of measurements (picomoles)") +
    ylab("Difference (picomoles)")
  return(BA)

}

#' run spike_counts on BAM/CRAM files and shape the results for model_glm_pmol
#'
#' Typically one will want to fit a correction model to multiple samples.
#' This function eases this task by merging the output of spike_counts into
#' a data.frame that model_glm_pmol can directly fit.
#'
#' @param files   a vector of BAM/CRAM file names
#' @param spike   a spike-in database
#' @param sep     the separator for spike-in contig names ("_")
#' @param methylated  a logical (0/1) to include only methylated fragments
#'
#' @return        a data.frame with columns "frag_grp", "id", and "read_count"
#'
#' @examples
#' data(spike)
#' library(GenomicRanges)
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                   mustWork=TRUE)
#' scan_spike_counts(sb, spike=spike)
#' fit <- model_glm_pmol(scan_spike_counts(sb, spike=spike),spike=spike)
#'
#' @export
scan_spike_counts <- function(files, spike, methylated=1,sep="_") {

  stopifnot(all(file.exists(files)))
  ssc <- lapply(files, spike_counts, spike=spike, dump_idx=TRUE)
  do.call(rbind, lapply(lapply(ssc, subset, methylated==methylated), .tidysc))

}


# helper fn
.tidysc <- function(sc) {

  sc$frag_grp <- .tidyfg(rownames(sc))
  sc$read_count <- sc$mapped
  sc[, c("frag_grp", "id", "read_count")]

}


# helper fn
.tidyfg <- function(frags) {

  gsub("[a-z]", "", tolower(sapply(strsplit(frags, "\\-"), `[`, 1)))

}

#' tabulate read coverage in predefined bins
#'
#' refactored out of scan_spiked_bam
#'
#' @param bins    the GRanges with bins
#' @param covg    the coverage result (an RleList)
#'
#' @return        a GRanges of summarized coverage
#'
#' @seealso       get_spiked_coverage
#' @seealso       scan_spiked_bam
#'
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky",
#'                    mustWork=TRUE)
#' data(spike,package="spiky")
#' si <- seqinfo_from_header(sb)
#' genome(si) <- "spike"
#' mgr <- get_merged_gr(si,spike=spike)
#' fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
#' bp <- ScanBamParam(flag=fl)
#' bamMapqFilter(bp) <- 20
#'
#' covg <- get_spiked_coverage(sb, bp=bp, gr=mgr)
#' get_binned_coverage(bins=GRanges(), covg=covg)
#'
#' @import        GenomicRanges
#'
#' @export
get_binned_coverage <- function(bins, covg) {

  if (length(bins) < 1) {
    message("Empty bins provided, skipping.")
    return(bins)
  }
  message("Binning genomic coverage...", appendLF=FALSE)
  bc <- binnedAverage(bins, numvar=covg[seqlevels(bins)], varname="coverage")
  message("Done.")
  return(bc)

}

#' Tile the assembly-based contigs of a merged assembly/spike GRanges.
#'
#' refactored out of scan_spiked_bam for more explicit information flow
#'
#' @param gr          the GRanges
#' @param binwidth    bin width to tile (default is 300)
#'
#' @return            a GRanges of bins
#'
#' @examples
#'
#' bam <- system.file("extdata", "ex1.bam", package="Rsamtools",
#'                    mustWork=TRUE)
#' gr <- as(seqinfo_from_header(bam), "GRanges")
#' genome(gr) <- "notspike"
#' tile_bins(gr)
#'
#'
#' @export
tile_bins <- function(gr, binwidth=300L) {

  chroms <- subset(seqlevels(gr), genome(gr)[seqlevels(gr)] != "spike" )
  si <- seqinfo(gr)[chroms]
  message("Tiling ", binwidth, "bp bins across the genome...", appendLF=FALSE)
  bins <- tileGenome(si, tilewidth=binwidth, cut.last.tile.in.chrom=TRUE)
  message("Done.")
  return(bins)

}

#' get a GRanges of (by default, standard) chromosomes from seqinfo
#' 
#' refactored from scan_spiked_bam to clarify information flow
#' 
#' @param si        seqinfo, usually from a BAM/CRAM file with spike contigs
#' @param spike     database of spike-in standard sequence features (spike)
#' @param standard  trim to standard chromosomes? (TRUE) 
#' 
#' @return          GRanges with two genomes: the organism assembly and "spike"
#'
#' @examples
#' sb <- system.file("extdata", "example.spike.bam", package="spiky", 
#'                   mustWork=TRUE) 
#' si <- seqinfo_from_header(sb) 
#' genome(si) <- "spike" # no genomic contigs
#' data(spike, package="spiky")
#' get_merged_gr(si, spike=spike) # note canonicalized spikes
#' 
#' @details
#' By default, get_merged_gr will return a GRanges with "standardized" 
#' genomic and spike contig names (i.e. genomic chr1-22, X, Y, M, and 
#' the canonical spike names in data(spike, package="spiky")).
#' 
#' The constraint to "standard" chromosomes on genomic contigs can be 
#' removed by setting `standard` to FALSE in the function arguments.
#'
#' @import          GenomeInfoDb
#' 
#' @export
get_merged_gr <- function(si, spike, standard=TRUE) { 

  # assembly contigs 
  if (standard) {
    chrom_contigs <- seqlevels(keepStandardChromosomes(si))
  } else {
    chrom_contigs <- seqlevels(si)
  }

  # spike contigs using the names found in the merged BAM/CRAM file
  orig_spike_contigs <- subset(seqlevels(si), genome(si) == "spike")

  # the collection of contigs we want to keep for downstream analysis
  merged_contigs <- sortSeqlevels(si[c(chrom_contigs, orig_spike_contigs)])

  # coerce this to a GRanges object 
  gr <- as(merged_contigs, "GRanges")

  # use standardized element names, but keep the seqlevels found in the BAM:
  names(gr) <- as.character(seqnames(rename_spike_seqlevels(gr, spike=spike)))
  return(gr)

} 

#' A handful of methods that I've always felt were missing
#'
#' Particularly, simple methods to plot coverage results.
#' 
#' @param  x                an Rle or RleList, usually
#' @param  y                not usedan Rle or RleList, usually
#' @param  ...              other params such as `ylim` passed to `barplot`
#' 
#' @return                  invisibly, the plot details
#' 
#' @details
#' selectMethod("plot", "Rle") and also
#' selectMethod("plot", "RleList") too.
#' 
#' @importFrom graphics barplot par text title
#' 
#' @rdname spiky-methods
#' @name   spiky-methods
NULL


#' @rdname spiky-methods
#' @export 
setMethod("plot", "Rle", 
          function(x, y, ...) {
            titlex <- sum(width(x)) * 0.9
            titley <- ifelse(exists("ylim"), ylim[2] * 0.9, max(x) * 0.9)
            barplot(runValue(x), width(x), las=2, inside=FALSE, col="blue", ...)
            if (exists("title")) text(titlex, titley, title)
          })


#' @rdname spiky-methods
#' @export 
setMethod("plot", "SimpleRleList", 
          function(x, y, ...) {
            yl <- c(0, max(max(x)))
            idx <- which(!all(x == 0))
            par(mfrow=c(length(idx), 1))
            par("mai"=c(0.05, 0.5, 0.05, 0.05))
            if (is.null(names(idx))) names(idx) <- paste0("run", idx)
            for (ni in names(idx)) plot(x[[idx[ni]]], ylim=yl, title=ni)
          })

#' read a BEDPE file into Pairs of GRanges (as if a GAlignmentPairs or similar)
#'
#' @param x         a Tabixed BEDPE file, or a TabixFile of one
#' @param ...       additional arguments to pass to scanTabix internally
#' @param optional  scan the optional columns (name, score, strand1)? (FALSE)
#' @param fraglen   compute the fragment length? (TRUE)
#' @param keep      keep additional columns? (FALSE)
#' @param stranded  Is the data stranded? (FALSE)
#'
#' @return          a Pairs of GRanges, perhaps with $score or $fraglen
#'
#' @details         BEDPE import in R is a shambles. This is a bandaid on a GSW.
#'                  See the \href{https://bedtools.readthedocs.io/en/latest/content/general-usage.html#bedpe-format}{BEDPE format definition} for full details.
#'                  In short, for a pair of ranges 1 and 2, we have fields
#'                  chrom1, start1, end1, chrom2, start2, end2, and (optionally)
#'                  name, score, strand1, strand2, plus any other user defined
#'                  fields that may be included (these are not yet supported
#'                  by read_bedpe). For example, two valid BEDPE lines are:
#'
#'                  chr1  100   200   chr5  5000  5100  bedpe_example1  30
#'                  chr9  900  5000   chr9  3000  3800  bedpe_example2  99  +  -
#'
#' @examples
#' \dontrun{
#'   bedpe <- "GSM5067076_2020_A64_bedpe.bed.gz"
#'   WT1_hg38 <- GRanges("chr11", IRanges(32387775, 32435564), "-")
#'   read_bedpe(bedpe, param=WT1_hg38)
#' }
#'
#' @seealso         bedpe_covg
#'
#' @import          Rsamtools
#' @import          S4Vectors
#'
#' @export
read_bedpe <- function(x, ..., stranded=FALSE, fraglen=TRUE, optional=FALSE, keep=FALSE) {

  # no support yet for additional mcols, even though it's kind of easy
  if (keep) message("Extra columns are not yet supported.")

  # mandatory for proper scanning and param support
  if (!is(x, "TabixFile")) x <- TabixFile(x)

  # format specification for BEDPE
  fmt <- c(seqnames1="character",
           start1="integer",
           end1="integer",
           seqnames2="character",
           start2="integer",
           end2="integer",
           name="name",
           score="numeric",
           strand1="character",
           strand2="character")

  # drop strand fields if the BEDPE is not stranded
  if (!stranded) fmt <- fmt[!grepl("strand[12]", names(fmt))]

  # switch to scanTabix -> parsing
  bits <- list(
    first=fmt[grep(1, names(fmt))],
    second=fmt[grep(2, names(fmt))],
    score=fmt[grep("score", names(fmt))]
  )

  # switch to scanTabix? this is slow AF as implemented
  tbx <- unlist(scanTabix(x, ...), use.names=FALSE)
  if (length(strsplit(tbx[1], "\t")[[1]]) < length(fmt)) { fmt <- fmt[!names(fmt)=="name"]}
  if (length(strsplit(tbx[1], "\t")[[1]]) < length(fmt)) {stop("Missing columns in input file.")}
  res <- lapply(bits, .parse_tabix, fmt=fmt, tbx=tbx)
  pe <- .toPairs(res)

  if (fraglen) mcols(pe)$fraglen <- (end(second(pe)) - start(first(pe)))
  if (all(is.na(mcols(pe)$score)) | all(mcols(pe)$score == 1)) mcols(pe)$score <- NULL

  return(pe)

}


# helper
.parse_tabix <- function(bits, fmt, tbx) {

  # field positions
  pos <- vapply(names(bits), grep, x=names(fmt), integer(1))
  dat <- do.call(rbind, strsplit(tbx, "\t"))[, pos]

  if (length(bits) > 1) {
    colnames(dat) <- sub("[12]$", "", names(bits))
    if (all(any(grepl("seqnames", names(bits))),
            any(grepl("start", names(bits))),
            any(grepl("end", names(bits))))) {
      dat <- makeGRangesFromDataFrame(data.frame(dat))
    }
  }

  return(dat)

}


# helper
.rename <- function(x, suffix="[12]$") {

  names(x) <- sub(suffix, "", names(x))
  return(x)

}


# bit of a hack utility
.toGRs <- function(x) {

  cols <- list(first=grep("(chrom|seqnames|start|end)1", colnames(x), val=TRUE),
               second=grep("(chrom|seqnames|start|end)2", colnames(x),val=TRUE))
  lapply(cols, function(y) makeGRangesFromDataFrame(.rename(x[, y])))

}


# helper
.toPairs <- function(res) {

  paired <- with(res, Pairs(first=first, second=second))
  for (extracol in setdiff(names(res), c("first", "second"))) {
    if (extracol == "score") mcols(paired)$score <- as.numeric(res$score)
    else mcols(paired)[, extracol] <- res[[extracol]]
  }
  return(paired)

}

---
title: "Spiky: Analysing cfMeDIP-seq data with spike-in controls"
author: "Samantha L Wilson and Lauren M Harmon"
date: "February 8, 2021"
output: rmarkdown::html_vignette
bibliography: references.bib 
csl: nature.csl
link-citations: yes
vignette: >
  %\VignetteIndexEntry{Spiky: Analysing cfMeDIP-seq data with spike-in controls}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(devtools)
load_all("./")
```

# Introduction

To meet the need for a reference control in cell-free methylated DNA immunoprecipitation-sequencing (cfMeDIP-seq)[@shen2018sensitive; @shen2019preparation] experiments, we designed spike-in controls and ligated unique molecular indexes (UMI) to adjust for PCR bias, and immunoprecipitation bias caused by the fragment length, G+C content, and CpG density of the DNA fragments that are immunoprecipitated[@wilson2022spikeins]. This enables absolute quantification of methylated DNA in picomoles, while retaining epigenomic information that allows for sensitive, tissue-specific detection as well as comparable results between different experiments. We designed DNA fragments with 2x3x3x3=54 combinations of methylation status (methylated and unmethylated), fragment length in basepair (bp) (80 bp, 160 bp, 320 bp), G+C content (35%, 50%, 65%), and fraction of CpGs within a fragment (1 CpG/ 80 bp, 1 CpG/ 40 bp, 1 CpG/ 20 bp). Spiky was developed for analyzing DNA methylation of cell-free DNA obtained from cfMeDIP-seq method using reference 'spike-in' controls. This package will:

* Assess methylation specificity in each sample
* Using the spike-in control data, output a Gaussian generalized linear model to predict molar amount on DNA samples
* Predict molar amount (picomoles) for each DNA sequence of interest, adjusting for fragment length G+C content and CpG fraction
* Adjust molar amount and bin the fragments into genomic windows used in analyses

# Installation
Install and load the spiky package from Bioconductor.
```{r eval = FALSE,message=FALSE}
#To install this package, start R (version "3.6" or later) and enter:
  #if (!requireNamespace("BiocManager", quietly = TRUE))
  #  install.packages("BiocManager")
  #
  #BiocManager::install("spiky")

library(spiky)
```

# Load spike database, or create your own with process_spikes().

### Input: A Fasta file, GRanges, or dataframe of spike-in sequences, and a vector of booleans (0 or 1) describing whether each spike-in sequence is methylated. 

### Output: The output contains a DataFrame with the following columns:
  *sequence (DNAStringSet)
  *methylated (boolean)
  *CpGs (integer)
  *fmol (numeric)
  *molmass (numeric)
  *GCfrac (numeric)
  *OECpG (numeric)
  *conc (numeric)
  *frag_grp (character)
  
If you are using the same synthetic spike-in sequences as described in the manuscript, you may load the spike-in sequence database using: 
```{r}
data(spike)
```

If you are using custom spike-ins, you can create your own spike-in sequence database using the process_spikes() function, which accepts as input a Fasta file, GRanges, or dataframe, and a vector of booleans (0 or 1) describing whether each spike-in sequence is methylated. The input Fasta file can also be generated from BAM header info as follows:
```{r}
sb <- system.file("extdata", "example.spike.bam", package="spiky",              mustWork=TRUE)
outFasta <- paste(system.file("extdata", package="spiky", mustWork=TRUE),"/spike_contigs.fa",sep="")
show(generate_spike_fasta(sb, spike=spike,fa=outFasta))
```

The spike-in database can then be created with this input Fasta.
```{r}
spikes <- system.file("extdata", "spikes.fa", package="spiky", mustWork=TRUE)
spikemeth <- spike$methylated
process_spikes(spikes, spikemeth)
```


# Process the input files
Spiky supports input files in either BAM or BEDPE format. 

### BAM Input
BAM file in standard format (For full details about the BAM format, see https://samtools.github.io/hts-specs/SAMv1.pdf). The BAM must also have an accompanying index file, which can be created using samtools index ${filename.bam}. (http://www.htslib.org/doc/samtools-index.html)

# BAM required columns
  * BAM file
  * Columns:
    * chrom/contig: string 
    * position start: numeric
    * position end: numeric
    * read counts: integer
    * fragment length (bp): integer

### Output: The output objects will be used downstream in the analysis, including
  * genomic_coverage - A GRanges object showing the genomic coverage of the BAM reads
  * spikes_coverage - A GRanges object showing the coverage of the spikes.
  
```{r eval=TRUE}
genomic_bam_path <- system.file("extdata", "example_chr21.bam", package="spiky", mustWork=TRUE)
genomic_coverage <- scan_genomic_contigs(genomic_bam_path,spike=spike)
spike_bam_path <- system.file("extdata", "example.spike.bam", package="spiky", mustWork=TRUE)
spikes_coverage <- scan_spike_contigs(spike_bam_path,spike=spike)

```

### BEDPE Input
BEDPE file in standard format. For full details about the BEDPE format, see Bedtools documentation (https://bedtools.readthedocs.io/en/latest/content/general-usage.html#bedpe-format).  In short, for a pair of ranges 1 and 2, we have fields chrom1, start1, end1, chrom2, start2, end2, and (optionally) name, score, strand1, strand2, plus any other user defined fields that may be included (these are not yet supported by Spiky). For example, two valid BEDPE lines are: 

chr1 100 200 chr5 5000 5100 bedpe_example1 30 

chr9 900 5000 chr9 3000 3800 bedpe_example2 99 + - 

The BEDPE must also have an accompanying index file, which can be created using Bedtools, as in the example shown below, where ${file} represents the name of a BEDPE file. 

bedtools sort -i ${file} > sorted_${file} 
bgzip sorted_${file} 
tabix sorted_${file}.gz 

### Output: The output objects will be used downstream in the analysis, including
  * genomic_coverage - A GRanges object showing the genomic coverage of the BAM reads
  * spikes_coverage - A GRanges object showing the coverage of the spikes.
  
```{r}
genomic_bedpe_path <- system.file("extdata", "example_chr21_bedpe.bed.gz", package="spiky", mustWork=TRUE)
genomic_coverage <- scan_genomic_bedpe(genomic_bedpe_path,genome="hg38")
spike_bedpe_path <- system.file("extdata", "example_spike_bedpe.bed.gz", package="spiky", mustWork=TRUE)
spikes_coverage <- scan_spike_bedpe(spike_bedpe_path,spike=spike)
```
  
# Methylation specificity
For each combination of parameters, we designed two distinct spike-in sequences. One to be methylated and one to be unmethylated. The allows us to assess non-specific binding of the monoclonal antibody on a sample-by-sample basis. To calculate methylation specificity we take the number of methylated reads divided by the total number of reads. It is our recommendation that if methylation specificity is <0.98, then the sample should be flagged or removed from analysis as the cfMeDIP performed inadequately. 

This calculation is done by the 'methylation_specificity' function.

### Input: The output of the 'scan_spike_contigs' or 'scan_spike_bedpe' functions
  * spikes_coverage as produced previously
  * spike database as produced previously
  
### Output: methylation specificity mean and median
  * Mean and median of the percent of methylated sequences for each spike-in after cfMeDIP-seq has been performed

### Example
```{r eval=TRUE}
##Calculate methylation specificity
methyl_spec <- methylation_specificity(spikes_coverage,spike=spike)
print(methyl_spec)
```

# Fit a Gaussian model to predict the molar amount of DNA sequences
For each batch of samples, the coefficients used in the Gaussian generalized linear model will differ. The 'model_glm_pmol' will calculate these coefficients and output the model to be used to calculate molar amount (picomoles) on the user's DNA sequences of interest. We assume that all DNA sequences of interest are methylated after undergoing cfMeDIP-seq. As such, we build the Gaussian generalized linear model on only the methylated spike-in control fragments. A generated Bland-Altman plot will visualize how well the model performs.

### Input: The output of the 'scan_spike_contigs' or 'scan_spike_bedpe' functions
  * spikes_coverage as produced previously
  * spike database as produced previously

### Output: 
  * Gaussian generalized linear model with coefficients specific to samples used in input data; .rda file

### Example
```{r eval=TRUE}
## Build the Gaussian generalized linear model on the spike-in control data
gaussian_glm <- model_glm_pmol(spikes_coverage,spike=spike)
summary(gaussian_glm)
```

# Calculating molar amount on DNA sequences of interest
For the samples in which the Gaussian generalized linear model was built, we will calculate the molar amount (picomoles) for each DNA sequence of interest.

### Input: The output of the 'scan_genomic_contigs' or 'scan_genomic_bedpe' functions and the Gaussian generalized linear model
  * genomic_coverage as produced previously
  * gaussian_glm as produced previously
  
### Output: sample_pmol_data
  * Data frame
  * Columns: 
    * chrom: string
    * bin position start: numeric
    * bin position end: numeric
    * read counts: coverage of bin
    * fragment length (bp): integer
    * G+C content [0-1]: numeric
    * CpG number: numeric 
    * pmol: numeric
  
### Example
```{r eval=TRUE}
# Predict pmol concentration
# To select a genome other than hg38, use BSgenome::available.packages() to find valid BSgenome name
#library("BSgenome.Hsapiens.UCSC.hg38")
sample_data_pmol <- predict_pmol(gaussian_glm, genomic_coverage,bsgenome="BSgenome.Hsapiens.UCSC.hg38",ret="df")
head(sample_data_pmol,n=1)

```

# Adjusting molar amount to binned genomic windows
For our analyses, we binned the genome into 300 bp non-overlapping windows. We then look overlap between fragments in our data with each of the 300 bp genomic windows. We adjust the molar amount (picomoles) by a multiplier. This multiplier is the proportion of overlap between our fragment and the 300 bp window. This is done for every fragment in our sample.

### Input: output dataframe produced from predict_pmol
   * Example: sample_pmol_data as produced in previous step
  
### Output: sample_binned_data
  * Data frame
  * Columns: 
    * chrom: string
    * bin position start: numeric
    * bin position end: numeric
    * read counts: coverage of bin
    * fragment length (bp): integer
    * G+C content [0-1]: numeric
    * CpG number: numeric 
    * pmol: numeric
    * adjusted pmol: numeric
  
### Example
```{r eval=TRUE}
sample_binned_data <- bin_pmol(sample_data_pmol)
head(sample_binned_data,n=1)
```
### Session Info
```{r}
sessionInfo()
```


# References
\chaptermark{references.bib}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{spike}
\alias{spike}
\title{spike-in contig properties for Sam's cfMeDIP spikes}
\format{
A DataFrame object with
\describe{
\item{sequence}{contig sequence, as a DNAStringSet}
\item{methylated}{are the CpGs in this spike-in methylated? 0 or 1}
\item{CpGs}{number of CpG dinucleotides in the spike, from 1 to 16}
\item{fmol}{femtomolar concentration of the spike-in for standard mix}
\item{molmass}{molar mass of spike-in sequence}
}
}
\source{
https://doi.org/10.1101/2021.02.12.430289
}
\usage{
data(spike)
}
\description{
A DataFrame with sequence, concentration, and other properties of Sam's
synthetic cfMeDIP spike-in controls. The row names redudantly encode some
of these properties, such as the number of CpGs in the spike-in sequence.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_base_name.R
\name{get_base_name}
\alias{get_base_name}
\title{refactored out of rename_spikes and rename_spike_seqlevels}
\usage{
get_base_name(contig_names, sep = "_")
}
\arguments{
\item{contig_names}{the names of contigs}

\item{sep}{separator character in contig names ("_")}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{               a vector of elements 1:3 from each contig name
}\if{html}{\out{</div>}}
}
\description{
A common task between generate_spike_fasta, rename_spikes, and
rename_spike_seqlevels is to determine what the largest common subset of
characters between existing contig names and stored standardized contigs
might be. This function eases that task.
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
bh <- scanBamHeader(BamFile(sb))
orig_contigs <- names(bh$targets)
get_base_name(orig_contigs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{spike_res}
\alias{spike_res}
\title{A Granges object with spike-in sequence coverage, and
summarized for each spike contig as (the default) \code{max} coverage. (In other
words, the default output of scan_spike_contigs or scan_spike_bedpe)
This represents what most users will want to generate from their own spike-in BAMs or BEDPEs, and
is used repeatedly in downstream examples throughout the package.}
\format{
A GRanges of coverage results with one metadata column, \code{coverage}
}
\source{
Generated using scan_spike_bedpe or scan_spike_contigs on an example bedpe or bam
containing spike contigs.
}
\usage{
data(spike_res)
}
\description{
A Granges object with spike-in sequence coverage, and
summarized for each spike contig as (the default) \code{max} coverage. (In other
words, the default output of scan_spike_contigs or scan_spike_bedpe)
This represents what most users will want to generate from their own spike-in BAMs or BEDPEs, and
is used repeatedly in downstream examples throughout the package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bam_to_bins.R
\name{bam_to_bins}
\alias{bam_to_bins}
\title{create a tiled representation of a genome from the BAM/CRAM file}
\usage{
bam_to_bins(x, width = 300, param = NULL, which = IRangesList(), ...)
}
\arguments{
\item{x}{a BAM or CRAM filename (or a BamFile object)}

\item{width}{the width of the bins to tile (default is 300)}

\item{param}{optional ScanBamParam (whence we attempt to extract \code{which})}

\item{which}{an optional GRanges restricting the bins to certain locations}

\item{...}{additional arguments to pass on to seqinfo_from_header}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   a GRangesList with y-base-pair-wide bins tiled across it
}\if{html}{\out{</div>}}
}
\description{
This function replaces a bedtools call:
bedtools intersect -wao -a fragments.bed -b hg38_300bp_windows.bed > data.bed
}
\details{
The idea is to skip the BED creation step for most runs, and just do it once.
In order to count reads in bins, we need bins.
In order to have bins, we need to know how long the chromosomes are.
In order to have a BAM or CRAM file, we need to have those same lengths.
This function takes advantage of all of the above to create binned ranges.
Note that a very recent branch of Rsamtools is required for CRAM file bins.
}
\examples{
library(Rsamtools) 
fl <- system.file("extdata", "ex1.bam", package="Rsamtools", mustWork=TRUE)
bam_to_bins(fl) 

}
\seealso{
seqinfo_from_header
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_spike_depth.R
\name{get_spike_depth}
\alias{get_spike_depth}
\title{get the (max, median, or mean) coverage for spike-in contigs from a BAM/CRAM}
\usage{
get_spike_depth(covg, spike_gr = NULL, spike = NULL, how = c("max", "mean"))
}
\arguments{
\item{covg}{the coverage RleList}

\item{spike_gr}{the spike-in GRanges (default: figure out from seqinfo)}

\item{spike}{information about the spikes (default: load \code{spike})}

\item{how}{how to summarize the per-spike coverage (max)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a GRanges with summarized coverage and features for each
}\if{html}{\out{</div>}}
}
\description{
get the (max, median, or mean) coverage for spike-in contigs from a BAM/CRAM
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
data(spike, package="spiky")
si <- seqinfo_from_header(sb)
genome(si) <- "spike"
mgr <- get_merged_gr(si,spike=spike)

fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
bp <- ScanBamParam(flag=fl)
bamMapqFilter(bp) <- 20

covg <- get_spiked_coverage(sb, bp=bp, gr=mgr)
get_spike_depth(covg, spike_gr=mgr, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kmers.R
\name{kmers}
\alias{kmers}
\title{oligonucleotideFrequency, but less letters and more convenient.}
\usage{
kmers(x, k = 6)
}
\arguments{
\item{x}{BSgenome, DFrame with \code{sequence} column, or DNAStringSet}

\item{k}{the length of the kmers (default is 6)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{       a matrix of contigs (rows) by kmer frequencies (columns)
}\if{html}{\out{</div>}}
}
\description{
oligonucleotideFrequency, but less letters and more convenient.
}
\details{
The companion \code{kmax} function finds the maximum frequency kmer for each
contig and plots all of them together for comparison purposes.
}
\examples{

data(genbank_mito, package="spiky") 
mtk6 <- kmers(genbank_mito, k=6)
kmax(mtk6)

data(phage, package="spiky") 
phk6 <- kmers(phage, k=6)
kmax(phk6)

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{      kmax
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_merged_gr.R
\name{get_merged_gr}
\alias{get_merged_gr}
\title{get a GRanges of (by default, standard) chromosomes from seqinfo}
\usage{
get_merged_gr(si, spike, standard = TRUE)
}
\arguments{
\item{si}{seqinfo, usually from a BAM/CRAM file with spike contigs}

\item{spike}{database of spike-in standard sequence features (spike)}

\item{standard}{trim to standard chromosomes? (TRUE)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     GRanges with two genomes: the organism assembly and "spike"
}\if{html}{\out{</div>}}
}
\description{
refactored from scan_spiked_bam to clarify information flow
}
\details{
By default, get_merged_gr will return a GRanges with "standardized"
genomic and spike contig names (i.e. genomic chr1-22, X, Y, M, and
the canonical spike names in data(spike, package="spiky")).

The constraint to "standard" chromosomes on genomic contigs can be
removed by setting \code{standard} to FALSE in the function arguments.
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky", 
                  mustWork=TRUE) 
si <- seqinfo_from_header(sb) 
genome(si) <- "spike" # no genomic contigs
data(spike, package="spiky")
get_merged_gr(si, spike=spike) # note canonicalized spikes

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_spike_contigs.R
\name{find_spike_contigs}
\alias{find_spike_contigs}
\title{find spike-in seqlevels in an object \code{x}, where !is.null(seqinfo(x))}
\usage{
find_spike_contigs(x, spike)
}
\arguments{
\item{x}{something with seqlevels}

\item{spike}{a DataFrame with spike-in information}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   indices of which contigs in seqlevels(x) are spike-in contigs
}\if{html}{\out{</div>}}
}
\description{
Find the spike-like contigs in a BAM with both natural and spiked contigs.
This started out as glue in some other functions and got refactored out.
}
\details{
The indices have an attribute "mappings", which is a character vector
such that attr(find_spike_contigs(x), "mappings") == standardized
for all contig names in the CRAM/BAM/whatever, and standardized is
the rowname in \code{spike} that corresponds to the original contig name.
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
si <- seqinfo_from_header(sb)
data(spike, package="spiky")
find_spike_contigs(si, spike=spike)

}
\seealso{
get_base_name

rename_spike_seqlevels
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covg_to_df.R
\name{covg_to_df}
\alias{covg_to_df}
\title{reshape \code{scan_spiked_bam} results into data.frames for model_glm_pmol}
\usage{
covg_to_df(spike_gr, spike, meth = TRUE, ID = NULL)
}
\arguments{
\item{spike_gr}{GRanges of spike contigs (e.g. output object from scan_spiked_bam, scan_spike_contigs, or scan_spike_bedpe)}

\item{spike}{spike database (as from data(spike, package="spiky"))}

\item{meth}{only keep methylated spike reads? (TRUE; if FALSE, sum both)}

\item{ID}{an identifier for this sample, if running several (autogenerate)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{ a data.frame with columns 'frag_grp', 'id', and 'read_count'
}\if{html}{\out{</div>}}
}
\description{
reshape \code{scan_spiked_bam} results into data.frames for model_glm_pmol
}
\examples{
data(spike, package="spiky")
data(spike_res, package="spiky")
subsetted <- covg_to_df(spike_res, spike=spike, meth=TRUE)
summed <- covg_to_df(spike_res, spike=spike, meth=FALSE)
round((summed$read_count - subsetted$read_count) / summed$read_count, 3)

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{scan_spiked_bam
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bin_pmol.R
\name{bin_pmol}
\alias{bin_pmol}
\title{Binned estimation of picomoles of DNA present in cfMeDIP assays}
\usage{
bin_pmol(x)
}
\arguments{
\item{x}{results from predict_pmol (a data.frame or GRanges)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{ the same object, but with a column `adjusted_pred_con`
}\if{html}{\out{</div>}}
}
\description{
Given the results of model_glm_pmol and predict_pmol, adjust the predictions
to reflect picomoles of captured DNA overlapping a given bin in the genome.
}
\examples{
data(spike, package="spiky")
data(spike_res, package="spiky")
data(genomic_res,package="spiky")
fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
pred <- predict_pmol(fit, genomic_res, ret="df")
bin_pmol(pred)

}
\seealso{
model_glm_pmol

predict_pmol
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{spike_cram_counts}
\alias{spike_cram_counts}
\title{spike-in counts, as a long data.frame}
\format{
A data.frame object with
\describe{
\item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
\item{id}{subject from whom cfMeDIP spike reads (column 3) were counted}
\item{read_count}{read coverage for this spike in this subject (column 2)}
}
}
\source{
Generated from scan_spike_counts(CRAMs, spike=spike) using example CRAMs
containing spike contigs
}
\usage{
data(spike_cram_counts)
}
\description{
A data.frame with spike-in results from CRAM files (generated from scan_spike_counts(CRAMs, spike=spike))
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_spikes.R
\name{rename_spikes}
\alias{rename_spikes}
\title{for BAM/CRAM files with renamed contigs, we need to rename \code{spike} rows}
\usage{
rename_spikes(x, spike)
}
\arguments{
\item{x}{a BAM/CRAM file, hopefully with an index}

\item{spike}{a DataFrame where spike$sequence is a DNAStringSet}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a DataFrame with renamed contigs (rows)
}\if{html}{\out{</div>}}
}
\description{
This function does that.
}
\seealso{
generate_spike_fasta
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_spikes.R
\name{process_spikes}
\alias{process_spikes}
\title{QC, QA, and processing for a new spike database}
\usage{
process_spikes(fasta, methylated = 0, ...)
}
\arguments{
\item{fasta}{fasta file (or GRanges or DataFrame) w/spike sequences}

\item{methylated}{whether CpGs in each are methylated (0 or 1, default 0)}

\item{...}{additional arguments, e.g. kernels (currently unused)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{         a DataFrame suitable for downstream processing
}\if{html}{\out{</div>}}
}
\description{
Sequence feature verification: never trust anyone, least of all yourself.
}
\details{
GCfrac is the GC content of spikes as a proportion instead of a percent.
OECpG is (observed/expected) CpGs (expectation is 25\% of GC dinucleotides).
}
\examples{

data(spike)
spikes <- system.file("extdata", "spikes.fa", package="spiky", mustWork=TRUE)
spikemeth <- spike$methylated
process_spikes(spikes, spikemeth)

data(phage)
phages <- system.file("extdata", "phages.fa", package="spiky", mustWork=TRUE)
identical(process_spikes(phage), phage)
identical(phage, process_spikes(phage))

data(genbank_mito)
(mt <- process_spikes(genbank_mito)) # see also genbank_mito.R
gb_mito <- system.file("extdata", "genbank_mito.R", package="spiky")


}
\seealso{
kmers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_glm_pmol.R
\name{model_glm_pmol}
\alias{model_glm_pmol}
\title{Build a generalized linear model from spike-ins to correct bias in cfMeDIP}
\usage{
model_glm_pmol(x, spike, conc = NULL, ...)
}
\arguments{
\item{x}{data w/frag_grp, id, and read_count; or scan_spiked_bam result}

\item{spike}{spike database, e.g. \code{data(spike, package='spiky')}}

\item{conc}{concentration for each spike (will be referenced if NULL)}

\item{...}{other arguments to pass to \code{glm} (e.g. \code{family})}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   the model fit for the data
}\if{html}{\out{</div>}}
}
\description{
formerly '2020_model_glm_fmol'.  Note that everything in x can be had from
a BAM/CRAM with spike contigs named as frag_grp (len_CpGs_GC) in the index
and in fact that is what scan_spiked_bam now does.
}
\examples{

data(spike, package="spiky")

data(spike_read_counts, package="spiky")
fit1 <- model_glm_pmol(spike_read_counts, spike=spike)

data(spike_res) # scan_spiked_bam result
fit2 <- model_glm_pmol(spike_res, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spike_counts.R
\name{spike_counts}
\alias{spike_counts}
\title{use the index of a spiked BAM/CRAM file for spike contig coverage}
\usage{
spike_counts(
  bam,
  spike,
  sep = "_",
  ref = "spike",
  verbose = FALSE,
  dump_idx = FALSE
)
}
\arguments{
\item{bam}{the BAM or CRAM file (MUST HAVE AN INDEX)}

\item{spike}{a data.frame, DataFrame, or similar with spikes}

\item{sep}{separator character in contig names ("_")}

\item{ref}{reference name for spike genome ("spike")}

\item{verbose}{be verbose? (FALSE)}

\item{dump_idx}{dump the renamed idxstats to aggregate? (FALSE)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{               a GRanges of spike-in contig read counts
}\if{html}{\out{</div>}}
}
\description{
It dawned on me one day that we don't even have to bother reading the file
if we have an index for a spiked BAM/CRAM result, since any fragments that
map properly to the spike contigs are generated from synthetic templates.
This function takes an index and a spike database (usually a DataFrame) as
inputs and provides a rough coverage estimate over "rehabilitated" contig
names (i.e., canonicalized contigs mapping to the database) as its output.
}
\details{
The argument \code{spike} has no default since we are attempting to refactor the
spike-in databases into their own data packages and allow more general use.
}
\examples{
data(spike, package="spiky")
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE) 
spike_counts(sb, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methylation_specificity.R
\name{methylation_specificity}
\alias{methylation_specificity}
\title{compute methylation specificity for spike-in standards}
\usage{
methylation_specificity(spike_gr, spike)
}
\arguments{
\item{spike_gr}{GRanges of spike contigs (e.g. output object from scan_spiked_bam, scan_spike_contigs, or scan_spike_bedpe)}

\item{spike}{spike contig database, if needed (e.g. data(spike))}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{         list with median and mean coverage across spike contigs
}\if{html}{\out{</div>}}
}
\description{
In a cfMeDIP experiment, the yield of methylated fragments should be >95\%
(ideally 98-99\%) due to the nature of the assay.
}
\examples{
data(genomic_res)
data(spike_res)
data(spike, package="spiky")
methylation_specificity(spike_res, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{spiky-methods}
\alias{spiky-methods}
\alias{plot,Rle,ANY-method}
\alias{plot,SimpleRleList,ANY-method}
\title{A handful of methods that I've always felt were missing}
\usage{
\S4method{plot}{Rle,ANY}(x, y, ...)

\S4method{plot}{SimpleRleList,ANY}(x, y, ...)
}
\arguments{
\item{x}{an Rle or RleList, usually}

\item{y}{not usedan Rle or RleList, usually}

\item{...}{other params such as \code{ylim} passed to \code{barplot}}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{             invisibly, the plot details
}\if{html}{\out{</div>}}
}
\description{
Particularly, simple methods to plot coverage results.
}
\details{
selectMethod("plot", "Rle") and also
selectMethod("plot", "RleList") too.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_genomic_bedpe.R
\name{scan_genomic_bedpe}
\alias{scan_genomic_bedpe}
\title{Scan genomic BEDPE}
\usage{
scan_genomic_bedpe(
  bedpe,
  bin = TRUE,
  binwidth = 300L,
  bins = NULL,
  standard = TRUE,
  genome = "hg38"
)
}
\arguments{
\item{bedpe}{the BEDPE file path, or output from read_bedpe()}

\item{bin}{Bin reads? (TRUE)}

\item{binwidth}{width of the bins for chromosomal tiling (300)}

\item{bins}{a pre-tiled GRanges for binning coverage (NULL)}

\item{standard}{restrict non-spike contigs to "standard" chromosomes? (TRUE)}

\item{genome}{Name of genome (default hg38)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a GRanges with coverage
}\if{html}{\out{</div>}}
}
\description{
Scan genomic BEDPE
}
\examples{

fl <- system.file("extdata", "example_chr21_bedpe.bed.gz", package="spiky",mustWork=TRUE)
scan_genomic_bedpe(fl) # will warn user about spike contigs


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_spike_UMI.R
\name{parse_spike_UMI}
\alias{parse_spike_UMI}
\title{parse out the forward and reverse UMIs and contig for a BED/BAM}
\usage{
parse_spike_UMI(UMI, pos = NULL, seqs = NULL)
}
\arguments{
\item{UMI}{a vector of UMIs}

\item{pos}{optional vector of positions (else all are set to 1)}

\item{seqs}{optional vector of read sequences (else widths default to 96)}
}
\value{
a GRanges
}
\description{
parse out the forward and reverse UMIs and contig for a BED/BAM
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_spike_contigs.R
\name{scan_spike_contigs}
\alias{scan_spike_contigs}
\title{pretty much what it says: scan spike contigs from a BAM or CRAM file}
\usage{
scan_spike_contigs(bam, spike, how = "max", param = NULL, mc.cores = 16, ...)
}
\arguments{
\item{bam}{the BAM or CRAM filename, or a vector of such filenames}

\item{spike}{the spike-in reference database (e.g. data(spike))}

\item{how}{how to summarize the per-spike coverage (max)}

\item{param}{a ScanBamParam object, or NULL (will default to MAPQ=20 etc)}

\item{mc.cores}{Number of cores to run on (default 16)}

\item{...}{additional arguments to pass to scanBamFlag()}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a CompressedGRangesList with bin- and spike-level coverage
}\if{html}{\out{</div>}}
}
\description{
default workflow is
}
\details{
\enumerate{
\item scan spike contigs and count fragments per contig or per bin.
\item fit the appropriate model for adjusting genomic contigs based on spikes.
\item scan and adjust binned fragment tallies along genomic contigs per above.
}

scan_spike_contigs implements step 1.

If multiple BAM or CRAM filenames are provided, all indices will be
checked before attempting to run through any of the files.
}
\examples{
library(GenomicRanges)
data(spike, package="spiky")
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE) # switch to a CRAM
res <- scan_spike_contigs(sb, spike=spike) # use default ScanBamParam
summary(res)

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    Rsamtools::ScanBamParam
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{spike_read_counts}
\alias{spike_read_counts}
\title{spike-in counts, as a long data.frame}
\format{
A data.frame object with
\describe{
\item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
\item{id}{subject from whom cfMeDIP spike reads (column 3) were counted}
\item{read_count}{read coverage for this spike in this subject (column 2)}
}
}
\source{
This data was created using inst/script/loadDedup.R
}
\usage{
data(spike_read_counts)
}
\description{
A data.frame with spike-in results from control samples in the manuscript.
This maps 1:1 onto \code{dedup} using reshape2::melt.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertPairedGRtoGR.R
\name{convertPairedGRtoGR}
\alias{convertPairedGRtoGR}
\title{Convert Pairs to GRanges}
\usage{
convertPairedGRtoGR(pairs)
}
\arguments{
\item{pairs}{the Pairs object}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{       a GRanges
}\if{html}{\out{</div>}}
}
\description{
Convert Pairs to GRanges
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_methylation_specificity.R
\name{scan_methylation_specificity}
\alias{scan_methylation_specificity}
\title{tabulate methylation specificity for multiple spike-in BAM/CRAM files}
\usage{
scan_methylation_specificity(files, spike, sep = "_")
}
\arguments{
\item{files}{a vector of BAM/CRAM file names}

\item{spike}{a spike-in database}

\item{sep}{the separator for spike-in contig names ("_")}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   a matrix with columns "mean" and "median"
}\if{html}{\out{</div>}}
}
\description{
Methylation specificity is here defined as methylated_spike_covg/spike_covg
}
\examples{
data(spike) 
library(GenomicRanges)
sb <- system.file("extdata", "example.spike.bam", package="spiky", 
                  mustWork=TRUE)
scan_methylation_specificity(sb, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_spike_seqlevels.R
\name{rename_spike_seqlevels}
\alias{rename_spike_seqlevels}
\title{for spike-in contigs in GRanges, match to standardized spike seqlevels}
\usage{
rename_spike_seqlevels(x, spike = NULL)
}
\arguments{
\item{x}{something with seqlevels (GRanges, GAlignments, Seqinfo...)}

\item{spike}{a DataFrame where spike$sequence is a DNAStringSet (or NULL)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     x, but with standardized spike seqlevels and genomes
}\if{html}{\out{</div>}}
}
\description{
This function is essentially the opposite of rename_spikes, except that
it works well on GRanges/GAlignments from or for merged genome+spike BAMs.
If spike contigs are found, it will assign genome='spike' to those, while
changing the seqlevels to standardized names that match rownames(spike).
}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    rename_spikes
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{genomic_res}
\alias{genomic_res}
\title{A Granges object with genomic coverage from chr21q22,
binned every 300bp for the genomic contigs then averaged across the bin. (In other
words, the default output of scan_genomic_contigs or scan_genomic_bedpe,
restricted to a small enough set of genomic regions to be practical for examples.)
This represents what most users will want to generate from their own genomic BAMs or BEDPEs, and
is used repeatedly in downstream examples throughout the package.}
\format{
A GRanges of coverage results with one metadata column, \code{coverage}
}
\source{
Generated using scan_genomic_bedpe or scan_genomic_contigs on an example bedpe or bam
containing chr21q22 contigs.
}
\usage{
data(genomic_res)
}
\description{
A Granges object with genomic coverage from chr21q22,
binned every 300bp for the genomic contigs then averaged across the bin. (In other
words, the default output of scan_genomic_contigs or scan_genomic_bedpe,
restricted to a small enough set of genomic regions to be practical for examples.)
This represents what most users will want to generate from their own genomic BAMs or BEDPEs, and
is used repeatedly in downstream examples throughout the package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spike_bland_altman_plot.R
\name{spike_bland_altman_plot}
\alias{spike_bland_altman_plot}
\title{Bland-Altman plot for cfMeDIP spike standards}
\usage{
spike_bland_altman_plot(fit)
}
\arguments{
\item{fit}{a model fit, from predict_pmol (?)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{             a ggplot2 object
}\if{html}{\out{</div>}}
}
\description{
Bland-Altman plot for cfMeDIP spike standards
}
\examples{

data(spike_res)
data(spike, package="spiky")
fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
ba_plot <- spike_bland_altman_plot(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{phage}
\alias{phage}
\title{lambda and phiX phage sequences, sometimes used as spike-ins}
\format{
A DataFrame object with
\describe{
\item{sequence}{genome sequence, as a DNAStringSet}
\item{methylated}{whether CpGs are methylated, as an integer}
\item{CpGs}{the number of CpGs in the phage genome, as an integer}
\item{GCfrac}{the GC fraction of the phage genome, as a numeric}
\item{OECpG}{the observed / expected CpG fraction, as a numeric}
}
}
\source{
www.ncbi.nlm.nih.gov/genbank/
}
\usage{
data(phage)
}
\description{
A DataFrame with sequence, methylated, CpGs, GCfrac, and OECpG for phages
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_frag_info.R
\name{add_frag_info}
\alias{add_frag_info}
\title{decode fragment identifiers for spike-in standards}
\usage{
add_frag_info(x, frag_grp = "frag_grp", spike = NULL)
}
\arguments{
\item{x}{data.frame with a column of spike information (see above)}

\item{frag_grp}{column name for the spike contig information (\code{frag_grp})}

\item{spike}{optional database of spike-in properties (none)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     the data.frame x, augmented with metadata columns
}\if{html}{\out{</div>}}
}
\description{
given a vector of fragment identifiers like \verb{160_2_35} or \verb{80b_1C_35G-2},
encoded typically as lengthInBp_numberOfCpGs_GCpercent, and optionally a
database of spike-in sequences corresponding to those fragments, add those
columns to the source data (along with, if present in the database, other
metadata such as standard concentrations, GC fraction, etc.) and return i
an updated DataFrame.
}
\examples{
data(spike_cram_counts) 
data(spike, package="spiky") 
spike <- subset(spike, methylated == 1)
add_frag_info(spike_cram_counts, spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_spiked_coverage.R
\name{get_spiked_coverage}
\alias{get_spiked_coverage}
\title{tabulate coverage across assembly and spike contig subset in natural order}
\usage{
get_spiked_coverage(bf, bp, gr)
}
\arguments{
\item{bf}{the BamFile object}

\item{bp}{the ScanBamParam object}

\item{gr}{the GRanges with sorted seqlevels}
}
\value{
a list of Rles
}
\description{
FIXME: this is wicked slow, ask Herve if a faster version exists
}
\details{
Refactored from scan_spiked_bam, this is a very simple wrapper
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky", 
                  mustWork=TRUE)
si <- seqinfo_from_header(sb) 
genome(si) <- "spike"
data(spike, package="spiky")
mgr <- get_merged_gr(si, spike=spike) # note canonicalized spikes

fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
bp <- ScanBamParam(flag=fl)
bamMapqFilter(bp) <- 20
get_spiked_coverage(sb, bp=bp, gr=mgr)

}
\seealso{
scan_spiked_bam

coverage
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_spike_bedpe.R
\name{scan_spike_bedpe}
\alias{scan_spike_bedpe}
\title{Scan spikes BEDPE}
\usage{
scan_spike_bedpe(bedpe, spike, how = "max")
}
\arguments{
\item{bedpe}{the BEDPE file path, or output from read_bedpe()}

\item{spike}{information about the spikes (default: load \code{spike})}

\item{how}{how to summarize the per-spike coverage (max)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a GRanges with coverage
}\if{html}{\out{</div>}}
}
\description{
Scan spikes BEDPE
}
\examples{
data(spike, package="spiky")
fl <- system.file("extdata", "example_spike_bedpe.bed.gz", package="spiky",mustWork=TRUE)
scan_spike_bedpe(fl,spike=spike) # will warn user about spike contigs

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_bam_standards.R
\name{model_bam_standards}
\alias{model_bam_standards}
\title{Build a Bayesian additive model from spike-ins to correct bias in *-seq}
\usage{
model_bam_standards(x, conc = NULL, fm = NULL, ...)
}
\arguments{
\item{x}{data with assorted feature information (GCfrac, CpGs, etc)}

\item{conc}{concentration for each spike (must be provided!)}

\item{fm}{model formula (conc ~ read_count + fraglen + GCfrac + CpGs_3)}

\item{...}{other arguments to pass to \code{bamlss}}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   the model fit for the data
}\if{html}{\out{</div>}}
}
\description{
Build a Bayesian additive model from spike-ins to correct bias in *-seq
}
\examples{

library(bamlss)
data(spike_cram_counts,package="spiky")
data(spike,package="spiky")
scc <- add_frag_info(spike_cram_counts, spike=spike)
scc$conc <- scc$conc * 0.9 # adjust for dilution
scc$CpGs_3 <- scc$CpGs ^ (1/3)
fit0 <- model_bam_standards(scc,
                            fm=conc ~ read_count + fraglen)
fit1 <- model_bam_standards(scc,
                            fm=conc ~ read_count + fraglen + GCfrac + CpGs_3)
DIC(fit0, fit1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_binned_coverage.R
\name{get_binned_coverage}
\alias{get_binned_coverage}
\title{tabulate read coverage in predefined bins}
\usage{
get_binned_coverage(bins, covg)
}
\arguments{
\item{bins}{the GRanges with bins}

\item{covg}{the coverage result (an RleList)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   a GRanges of summarized coverage
}\if{html}{\out{</div>}}
}
\description{
refactored out of scan_spiked_bam
}
\examples{
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                   mustWork=TRUE)
data(spike,package="spiky")
si <- seqinfo_from_header(sb)
genome(si) <- "spike"
mgr <- get_merged_gr(si,spike=spike)
fl <- scanBamFlag(isDuplicate=FALSE, isPaired=TRUE, isProperPair=TRUE)
bp <- ScanBamParam(flag=fl)
bamMapqFilter(bp) <- 20

covg <- get_spiked_coverage(sb, bp=bp, gr=mgr)
get_binned_coverage(bins=GRanges(), covg=covg)

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{  get_spiked_coverage
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{  scan_spiked_bam
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tile_bins.R
\name{tile_bins}
\alias{tile_bins}
\title{Tile the assembly-based contigs of a merged assembly/spike GRanges.}
\usage{
tile_bins(gr, binwidth = 300L)
}
\arguments{
\item{gr}{the GRanges}

\item{binwidth}{bin width to tile (default is 300)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{       a GRanges of bins
}\if{html}{\out{</div>}}
}
\description{
refactored out of scan_spiked_bam for more explicit information flow
}
\examples{

bam <- system.file("extdata", "ex1.bam", package="Rsamtools",
                   mustWork=TRUE)
gr <- as(seqinfo_from_header(bam), "GRanges")
genome(gr) <- "notspike"
tile_bins(gr)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqinfo_from_header.R
\name{seqinfo_from_header}
\alias{seqinfo_from_header}
\title{create seqinfo (and thus a standard chromosome filter) from a BAM header}
\usage{
seqinfo_from_header(x, gen = NA, std = FALSE, ret = c("si", "gr"))
}
\arguments{
\item{x}{the BAM file or its header}

\item{gen}{genome of the BAM file, if known (NULL; autodetect)}

\item{std}{standard chromosomes only? (FALSE; will be empty if spikes)}

\item{ret}{return Seqinfo ("si", the default) or GRanges ("gr")? ("si")}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   Seqinfo object or GRanges (or `as(seqinfo, "GRanges")`)
}\if{html}{\out{</div>}}
}
\description{
create seqinfo (and thus a standard chromosome filter) from a BAM header
}
\details{
Setting std=TRUE on a spike-in BAM will produce an empty result.
}
\examples{

  library(Rsamtools)
  fl <- system.file("extdata", "ex1.bam", package="Rsamtools", mustWork=TRUE)

  hdr <- scanBamHeader(BamFile(fl))
  si <- seqinfo_from_header(hdr)
  gr <- seqinfo_from_header(fl, ret="gr")
  stopifnot(identical(gr, as(si, "GRanges")))

  std_si <- seqinfo_from_header(fl, std=TRUE)
  seqlevels(std_si)

  # for comparison with below
  data(spike, package="spiky")
  spike

  sp <- system.file("extdata", "example.spike.bam", package="spiky")
  sp_gr <- seqinfo_from_header(sp, ret="gr")
  sp_gr

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{genbank_mito}
\alias{genbank_mito}
\title{various mitochondrial genomes sometimes used as endogenous spike-ins}
\format{
A DataFrame object with
\describe{
\item{species}{the species whence the record came, as a character string}
\item{genome}{the genome assembly whence the mtDNA, as a character string}
\item{accession}{the genbank accession, as a character string}
\item{sequence}{genome sequence, as a DNAStringSet}
}
}
\source{
www.ncbi.nlm.nih.gov/genbank/
}
\usage{
data(genbank_mito)
}
\description{
A DataFrame with species, genome, accession, and sequence for GenBank
mitochondrial genome depositions. No concentration provided; add if needed.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_bedpe.R
\name{read_bedpe}
\alias{read_bedpe}
\title{read a BEDPE file into Pairs of GRanges (as if a GAlignmentPairs or similar)}
\usage{
read_bedpe(
  x,
  ...,
  stranded = FALSE,
  fraglen = TRUE,
  optional = FALSE,
  keep = FALSE
)
}
\arguments{
\item{x}{a Tabixed BEDPE file, or a TabixFile of one}

\item{...}{additional arguments to pass to scanTabix internally}

\item{stranded}{Is the data stranded? (FALSE)}

\item{fraglen}{compute the fragment length? (TRUE)}

\item{optional}{scan the optional columns (name, score, strand1)? (FALSE)}

\item{keep}{keep additional columns? (FALSE)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a Pairs of GRanges, perhaps with $score or $fraglen
}\if{html}{\out{</div>}}
}
\description{
read a BEDPE file into Pairs of GRanges (as if a GAlignmentPairs or similar)
}
\details{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    BEDPE import in R is a shambles. This is a bandaid on a GSW.
             See the \\href\{https://bedtools.readthedocs.io/en/latest/content/general-usage.html#bedpe-format\}\{BEDPE format definition\} for full details.
             In short, for a pair of ranges 1 and 2, we have fields
             chrom1, start1, end1, chrom2, start2, end2, and (optionally)
             name, score, strand1, strand2, plus any other user defined
             fields that may be included (these are not yet supported
             by read_bedpe). For example, two valid BEDPE lines are:

             chr1  100   200   chr5  5000  5100  bedpe_example1  30
             chr9  900  5000   chr9  3000  3800  bedpe_example2  99  +  -
}\if{html}{\out{</div>}}
}
\examples{
\dontrun{
  bedpe <- "GSM5067076_2020_A64_bedpe.bed.gz"
  WT1_hg38 <- GRanges("chr11", IRanges(32387775, 32435564), "-")
  read_bedpe(bedpe, param=WT1_hg38)
}

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    bedpe_covg
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_spike_counts.R
\name{scan_spike_counts}
\alias{scan_spike_counts}
\title{run spike_counts on BAM/CRAM files and shape the results for model_glm_pmol}
\usage{
scan_spike_counts(files, spike, methylated = 1, sep = "_")
}
\arguments{
\item{files}{a vector of BAM/CRAM file names}

\item{spike}{a spike-in database}

\item{methylated}{a logical (0/1) to include only methylated fragments}

\item{sep}{the separator for spike-in contig names ("_")}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{   a data.frame with columns "frag_grp", "id", and "read_count"
}\if{html}{\out{</div>}}
}
\description{
Typically one will want to fit a correction model to multiple samples.
This function eases this task by merging the output of spike_counts into
a data.frame that model_glm_pmol can directly fit.
}
\examples{
data(spike)
library(GenomicRanges)
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
scan_spike_counts(sb, spike=spike)
fit <- model_glm_pmol(scan_spike_counts(sb, spike=spike),spike=spike)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kmax.R
\name{kmax}
\alias{kmax}
\title{simple contig kmer comparisons}
\usage{
kmax(km, normalize = TRUE)
}
\arguments{
\item{km}{kmer summary}

\item{normalize}{normalize (divide by row sums)? (TRUE)}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{       the most common kmers for each contig, across all contigs
}\if{html}{\out{</div>}}
}
\description{
simple contig kmer comparisons
}
\examples{


data(genbank_mito, package="spiky")
mtk6 <- kmers(genbank_mito, k=6)
rownames(mtk6) <- paste0(rownames(mtk6), "_MT")
kmax(mtk6)

data(phage, package="spiky")
phk6 <- kmers(phage, k=6)
kmax(phk6, normalize=FALSE)

stopifnot(identical(colnames(phk6), colnames(mtk6)))
k6 <- rbind(mtk6, phk6)
kmax(k6)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_genomic_contigs.R
\name{scan_genomic_contigs}
\alias{scan_genomic_contigs}
\title{scan genomic contigs in a BAM/CRAM file}
\usage{
scan_genomic_contigs(
  bam,
  spike,
  param = NULL,
  bin = TRUE,
  binwidth = 300L,
  bins = NULL,
  standard = TRUE,
  genome = "hg38",
  ...
)
}
\arguments{
\item{bam}{the BAM or CRAM filename, or a vector of them}

\item{spike}{the spike-in reference database (e.g. data(spike))}

\item{param}{a ScanBamParam object specifying which reads to count (NULL)}

\item{bin}{Bin reads? (TRUE)}

\item{binwidth}{width of the bins for chromosomal tiling (300)}

\item{bins}{a pre-tiled GRanges for binning coverage (NULL)}

\item{standard}{restrict non-spike contigs to "standard" chromosomes? (TRUE)}

\item{genome}{Name of genome (default hg38)}

\item{...}{additional arguments to pass to scanBamFlag()}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a CompressedGRangesList with bin- and spike-level coverage
}\if{html}{\out{</div>}}
}
\description{
The default workflow for spiky is roughly as follows:
}
\details{
\enumerate{
\item Identify and quantify the spike-in contigs in an experiment.
\item Fit a model for sequence-based abundance artifacts using the spike-ins.
\item Quantify raw fragment abundance on genomic contigs, and adjust per step 2.
}

scan_genomic_contigs addresses the first half of step 3. The assumption is
that anything which isn't a spike contig, is a genomic contig.  This isn't
necessarily true, so the user can also supply a ScanBamParam object for the
\code{param} argument and restrict scanning to whatever contigs they wish, which
also allows for non-default MAPQ, pairing, and quality filters.

If multiple BAM or CRAM filenames are provided, all indices will be
checked before attempting to run through any of the files.
}
\examples{

library(Rsamtools)
data(spike, package="spiky")

fl <- system.file("extdata", "ex1.bam", package="Rsamtools",
                  mustWork=TRUE)
scan_genomic_contigs(fl, spike=spike,standard=FALSE) # will warn user about spike contigs

sb <- system.file("extdata", "example_chr21.bam", package="spiky",
                  mustWork=TRUE)
scan_genomic_contigs(sb, spike=spike) # will warn user about genomic contigs

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    Rsamtools::ScanBamParam
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_spiked_bam.R
\name{scan_spiked_bam}
\alias{scan_spiked_bam}
\title{pretty much what it says: scan standard chroms + spike contigs from a BAM}
\usage{
scan_spiked_bam(
  bam,
  spike,
  mapq = 20,
  binwidth = 300L,
  bins = NULL,
  how = c("max", "mean"),
  dupe = FALSE,
  paired = TRUE,
  standard = TRUE,
  ...
)
}
\arguments{
\item{bam}{the BAM file}

\item{spike}{the spike-in reference database (e.g. data(spike))}

\item{mapq}{minimum mapq value to count a pair (20)}

\item{binwidth}{width of the bins for chromosomal tiling (300)}

\item{bins}{a pre-tiled GRanges for binning coverage (NULL)}

\item{how}{how to record spike read coverage (max or mean)? (max)}

\item{dupe}{unique (FALSE), duplicte (TRUE), or all (NA) reads? (FALSE)}

\item{paired}{restrict coverage to that from properly paired reads? (TRUE)}

\item{standard}{restrict non-spike contigs to "standard" chromosomes? (TRUE)}

\item{...}{additional arguments to pass to scanBamFlag()}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     a CompressedGRangesList with bin- and spike-level coverage
}\if{html}{\out{</div>}}
}
\description{
Note: behind the scenes, this is being refactored into scan_spike_contigs
and scan_genomic_contigs. Once that is done, perhaps before release, the
default workflow will switch to
}
\details{
\enumerate{
\item scan spike contigs and count fragments per contig or per bin.
\item fit the appropriate model for adjusting genomic contigs based on spikes.
\item scan and adjust binned fragment tallies along genomic contigs per above.
}

This approach decouples binning schemes from model generation (using spikes)
and model-based adjustment (using genomic fragment counts), decreasing code
complexity while increasing the opportunities for caching & parallelization.

For a more realistic example (not run), one might do something like:

data(spike, package="spiky");
bam <- "2021_ctl.hg38_withSpikes.bam";
ssb_res <- scan_spiked_bam(bam, mapq=20, spike=spike);

An extract from the resulting \code{ssb_res} object is available via

data(ssb_res, package="spiky");

The full ssb_res is a GRangesList object with 300bp-binned coverage on the
standard (chr1-22, chrX, chrY, chrM) chromosomes (as determined by the
GenomeInfoDb::standardChromosomes() function against the assembly defined
in the BAM or CRAM file, by default; if desired, a user can scan all
genomic contigs by setting standard=FALSE when calling the function).
By default, the mean base-level coverage of genomic bins is reported,
and the maximum spike-level coverage is reported, though this can also
be adjusted as needed. The results then inform the reliability of
measurements from replicate samples in multiple labs, as well as the
adjusted quantitative coverage in each bin once the absolute quantity
of captured cell-free methylated DNA has been fit by model_glm_pmol and
predict_pmol. In some sense, this function converts BAMs/CRAMs into usable
data structures for high-throughput standardized cfMeDIP experiments.

The data extract used in other examples is the same as the full version,
with the sole difference being that genomic bins are limited to chr22.
}
\examples{
library(GenomicRanges)
data(spike, package="spiky")
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
res <- scan_spiked_bam(sb, spike=spike, bins=GRanges())
summary(res$spikes$coverage)

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    GenomeInfoDb::keepStandardChromosomes
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{    Rsamtools::ScanBamParam
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ssb_res}
\alias{ssb_res}
\title{scan_spiked_bam results from a merged cfMeDIP CRAM file (chr22 and spikes)}
\format{
A CompressedGRangesList of coverage results, containing
\describe{
\item{genomic}{a GRanges with one metadata column, \code{coverage}}
\item{spikes}{a GRanges with one metadata column, \code{coverage}}
}
}
\source{
Generated using scan_spiked_bam on an example bam
containing chr22 and spike contigs.
}
\usage{
data(ssb_res)
}
\description{
A CompressedGRangesList object with \code{genomic} (chr22) and \code{spikes} coverage,
binned every 300bp for the genomic contigs then averaged across the bin, and
summarized for each spike contig as (the default) \code{max} coverage. (In other
words, the default output of scan_spiked_bam, restricted to a small enough
set of genomic regions to be practical for examples.) This represents what
most users will want to generate from their own merged BAMs or CRAMs, and
is used repeatedly in downstream examples throughout the package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_spike_fasta.R
\name{generate_spike_fasta}
\alias{generate_spike_fasta}
\title{for CRAM files, a FASTA reference is required to decode; this builds that}
\usage{
generate_spike_fasta(bam, spike, assembly = NULL, fa = "spike_contigs.fa")
}
\arguments{
\item{bam}{a BAM or CRAM file, hopefully with an index}

\item{spike}{the spike contig database (mandatory as of 0.9.99)}

\item{assembly}{optional BSgenome or seqinfo with reference contigs (NULL)}

\item{fa}{the filename for the resulting FASTA ("spikes.fa")}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{     invisibly, a DNAStringSet as exported to `fa`
}\if{html}{\out{</div>}}
}
\description{
A FASTA reference is \emph{not} always needed, so long as .crai indices
are available for all contigs in the CRAM. See \code{spike_counts} for a fast
and convenient alternative that extracts spike coverage from index stats.
However, spike_counts has its own issues, and it's better to use fragments.
}
\details{
If the contigs in a CRAM have even slightly different names from those in
the reference, decoding will fail.  In some cases there are multiple names
for a given contig (which raises the question of whether to condense them),
and thus the same reference sequence decodes multiple contig names.

This function generates an appropriate spike reference for a BAM or CRAM,
using BAM/CRAM headers to figure out which references are used for which.

At the moment, CRAM support in Rsamtools only exists in the GitHub branch:

BiocManager::install("Bioconductor/Rsamtools@cram")

Using other versions of Rsamtools will yield an error on CRAM files.

Note that for merged genomic + spike reference BAMs/CRAMs, this function
will only attempt to generate a FASTA for the spike contigs, not reference.
If your reference contigs are screwed up, talk to your sequencing people,
and keep better track of the FASTA reference against which you compress!
}
\examples{

library(GenomicRanges)
data(spike, package="spiky")
sb <- system.file("extdata", "example.spike.bam", package="spiky",
                  mustWork=TRUE)
outFasta <- paste(system.file("extdata", package="spiky", mustWork=TRUE),"/spike_contigs.fa",sep="")
show(generate_spike_fasta(sb, spike=spike,fa=outFasta))

}
\seealso{
\if{html}{\out{<div class="sourceCode">}}\preformatted{    rename_contigs
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dedup}
\alias{dedup}
\title{spike-in counts for two samples, as a wide data.frame}
\format{
A data.frame object with
\describe{
\item{frag_grp}{the encoded spike contig name: basepairs_CpGs_GCpercent}
\item{read_count_6547}{read coverage for this spike in sample 6547}
\item{read_count_6548}{read coverage for this spike in sample 6548}
}
}
\source{
This data was created using inst/script/loadDedup.R
}
\usage{
data(dedup)
}
\description{
A data.frame with spike-in results from control samples in the manuscript.
This maps 1:1 onto \code{spike_read_counts} using reshape2::melt.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_pmol.R
\name{predict_pmol}
\alias{predict_pmol}
\title{predict picomoles of DNA from a fit and read counts (coverage)}
\usage{
predict_pmol(
  fit,
  genomic_gr,
  bsgenome = NULL,
  ret = c("gr", "df"),
  slide = FALSE
)
}
\arguments{
\item{fit}{result of model_glm_pmol}

\item{genomic_gr}{the genomic data / new data}

\item{bsgenome}{BSgenome name (if null, will guess from genomic_gr)}

\item{ret}{return a data.frame ("df") or GRanges ("gr")?  ("gr")}

\item{slide}{compute a sliding window estimate for GCfrac (1/3 width)?}
}
\value{
object with read count, fraglen, GC\%, CpG**(1/3), and concentration
}
\description{
FIXME: this could be made MUCH faster by precomputing CpG/GC stats per bin
}
\details{
Using GRanges as the return value is (perhaps counterintuitively) \emph{much}
faster than the data.frame, since the sequence of the bins gets converted
from a BSgenome representation to characters in the latter (it is implied
by the bin start, stop, and genome when left as a GRanges).
}
\examples{

data(spike_res)
data(genomic_res)
data(spike, package="spiky")
fit <- model_glm_pmol(covg_to_df(spike_res, spike=spike),spike=spike)
preddf <- predict_pmol(fit, genomic_res, ret="df")
pred <- predict_pmol(fit, genomic_res, ret="gr")
bin_pmol(pred)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{testGR}
\alias{testGR}
\title{a test GRanges with UMI'ed genomic sequences used as controls}
\format{
A GRanges object with an mcols() DataFrame containing
\describe{
\item{UMI1}{the unique molecular identifier on the forward read}
\item{UMI2}{the unique molecular identifier on the reverse read}
\item{seq}{the sequence of the fragment}
\item{name}{the name of the fragment}
\item{score}{whether the fragment passes filters (always 1)}
}
}
\source{
Generated using inst/script/loadTest.R
}
\usage{
data(testGR)
}
\description{
Sources and overlap widths of various read sequences in a test CRAM.
}
\keyword{datasets}
